# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import itertools
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import vmap
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from jaxley.modules.base import Module
from jaxley.modules.cell import Cell
from jaxley.utils.cell_utils import (
    build_branchpoint_group_inds,
    compute_children_and_parents,
    convert_point_process_to_distributed,
    loc_of_index,
    merge_cells,
)
from jaxley.utils.misc_utils import concat_and_ignore_empty, cumsum_leading_zero
from jaxley.utils.solver_utils import (
    JaxleySolveIndexer,
    comp_edges_to_indices,
    remap_index_to_masked,
)
from jaxley.utils.syn_utils import gather_synapes


class Network(Module):
    """Network class.

    This class defines a network of cells that can be connected with synapses.
    """

    network_params: Dict = {}
    network_states: Dict = {}

    def __init__(
        self,
        cells: List[Cell],
    ):
        """Initialize network of cells and synapses.

        Args:
            cells: A list of cells that make up the network.
        """
        super().__init__()
        for cell in cells:
            self.xyzr += deepcopy(cell.xyzr)

        self._cells_list = cells
        self.ncomp_per_branch = np.concatenate(
            [cell.ncomp_per_branch for cell in cells]
        )
        self.ncomp = int(np.max(self.ncomp_per_branch))
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)
        self._internal_node_inds = np.arange(self.cumsum_ncomp[-1])
        self._append_params_and_states(self.network_params, self.network_states)

        self.nbranches_per_cell = [cell.total_nbranches for cell in cells]
        self.total_nbranches = sum(self.nbranches_per_cell)
        self._cumsum_nbranches = cumsum_leading_zero(self.nbranches_per_cell)

        self.nodes = pd.concat([c.nodes for c in cells], ignore_index=True)
        self.nodes["global_comp_index"] = np.arange(self.cumsum_ncomp[-1])
        self.nodes["global_branch_index"] = np.repeat(
            np.arange(self.total_nbranches), self.ncomp_per_branch
        ).tolist()
        self.nodes["global_cell_index"] = list(
            itertools.chain(
                *[[i] * int(cell.cumsum_ncomp[-1]) for i, cell in enumerate(cells)]
            )
        )
        self._update_local_indices()
        self._init_view()

        parents = [cell.comb_parents for cell in cells]
        self.comb_parents = jnp.concatenate(
            [p.at[1:].add(self._cumsum_nbranches[i]) for i, p in enumerate(parents)]
        )

        # Two columns: `parent_branch_index` and `child_branch_index`. One row per
        # branch, apart from those branches which do not have a parent (i.e.
        # -1 in parents). For every branch, tracks the global index of that branch
        # (`child_branch_index`) and the global index of its parent
        # (`parent_branch_index`).
        self.branch_edges = pd.DataFrame(
            dict(
                parent_branch_index=self.comb_parents[self.comb_parents != -1],
                child_branch_index=np.where(self.comb_parents != -1)[0],
            )
        )

        # For morphology indexing of both `jax.sparse` and the custom `jaxley` solvers.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )

        # `nbranchpoints` in each cell == cell._par_inds (because `par_inds` are unique).
        nbranchpoints = jnp.asarray([len(cell._par_inds) for cell in cells])
        self._cumsum_nbranchpoints_per_cell = cumsum_leading_zero(nbranchpoints)

        # Channels.
        self._gather_channels_from_constituents(cells)

        self._initialize()
        del self._cells_list

    def __repr__(self):
        return f"{type(self).__name__} with {len(self.channels)} different channels and {len(self.synapses)} synapses. Use `.nodes` or `.edges` for details."

    def _init_morph_jaxley_spsolve(self):
        branchpoint_group_inds = build_branchpoint_group_inds(
            len(self._par_inds),
            self._child_belongs_to_branchpoint,
            self.cumsum_ncomp[-1],
        )
        children_in_level = merge_cells(
            self._cumsum_nbranches,
            self._cumsum_nbranchpoints_per_cell,
            [cell._solve_indexer.children_in_level for cell in self._cells_list],
            exclude_first=False,
        )
        parents_in_level = merge_cells(
            self._cumsum_nbranches,
            self._cumsum_nbranchpoints_per_cell,
            [cell._solve_indexer.parents_in_level for cell in self._cells_list],
            exclude_first=False,
        )
        padded_cumsum_ncomp = cumsum_leading_zero(
            np.concatenate(
                [np.diff(cell._solve_indexer.cumsum_ncomp) for cell in self._cells_list]
            )
        )

        # Generate mapping to dealing with the masking which allows using the custom
        # sparse solver to deal with different ncomp per branch.
        remapped_node_indices = remap_index_to_masked(
            self._internal_node_inds,
            self.nodes,
            padded_cumsum_ncomp,
            self.ncomp_per_branch,
        )
        self._solve_indexer = JaxleySolveIndexer(
            cumsum_ncomp=padded_cumsum_ncomp,
            ncomp_per_branch=self.ncomp_per_branch,
            branchpoint_group_inds=branchpoint_group_inds,
            children_in_level=children_in_level,
            parents_in_level=parents_in_level,
            root_inds=self._cumsum_nbranches[:-1],
            remapped_node_indices=remapped_node_indices,
        )

    def _init_morph_jax_spsolve(self):
        """Initialize the morphology for networks.

        The reason that this function is a bit involved for a `Network` is that Jaxley
        considers branchpoint nodes to be at the very end of __all__ nodes (i.e. the
        branchpoints of the first cell are even after the compartments of the second
        cell. The reason for this is that, otherwise, `cumsum_ncomp` becomes tricky).

        To achieve this, we first loop over all compartments and append them, and then
        loop over all branchpoints and append those. The code for building the indices
        from the `comp_edges` is identical to `jx.Cell`.

        Explanation of `self._comp_eges['type']`:
        `type == 0`: compartment <--> compartment (within branch)
        `type == 1`: branchpoint --> parent-compartment
        `type == 2`: branchpoint --> child-compartment
        `type == 3`: parent-compartment --> branchpoint
        `type == 4`: child-compartment --> branchpoint
        """
        self._cumsum_ncomp_per_cell = cumsum_leading_zero(
            jnp.asarray([cell.cumsum_ncomp[-1] for cell in self.cells])
        )
        self._comp_edges = pd.DataFrame()

        # Add all the internal nodes.
        for offset, cell in zip(self._cumsum_ncomp_per_cell, self._cells_list):
            condition = cell._comp_edges["type"].to_numpy() == 0
            rows = cell._comp_edges[condition]
            self._comp_edges = pd.concat(
                [self._comp_edges, [offset, offset, 0] + rows], ignore_index=True
            )

        # All branchpoint-to-compartment nodes.
        start_branchpoints = self.cumsum_ncomp[-1]  # Index of the first branchpoint.
        for offset, offset_branchpoints, cell in zip(
            self._cumsum_ncomp_per_cell,
            self._cumsum_nbranchpoints_per_cell,
            self._cells_list,
        ):
            offset_within_cell = cell.cumsum_ncomp[-1]
            condition = cell._comp_edges["type"].isin([1, 2])
            rows = cell._comp_edges[condition]
            self._comp_edges = pd.concat(
                [
                    self._comp_edges,
                    [
                        start_branchpoints - offset_within_cell + offset_branchpoints,
                        offset,
                        0,
                    ]
                    + rows,
                ],
                ignore_index=True,
            )

        # All compartment-to-branchpoint nodes.
        for offset, offset_branchpoints, cell in zip(
            self._cumsum_ncomp_per_cell,
            self._cumsum_nbranchpoints_per_cell,
            self._cells_list,
        ):
            offset_within_cell = cell.cumsum_ncomp[-1]
            condition = cell._comp_edges["type"].isin([3, 4])
            rows = cell._comp_edges[condition]
            self._comp_edges = pd.concat(
                [
                    self._comp_edges,
                    [
                        offset,
                        start_branchpoints - offset_within_cell + offset_branchpoints,
                        0,
                    ]
                    + rows,
                ],
                ignore_index=True,
            )

        # Convert comp_edges to the index format required for `jax.sparse` solvers.
        n_nodes, data_inds, indices, indptr = comp_edges_to_indices(self._comp_edges)
        self._n_nodes = n_nodes
        self._data_inds = data_inds
        self._indices_jax_spsolve = indices
        self._indptr_jax_spsolve = indptr

    def _step_synapse(
        self,
        states: Dict,
        syn_channels: List,
        params: Dict,
        delta_t: float,
        edges: pd.DataFrame,
    ) -> Tuple[Dict, Tuple[jnp.ndarray, jnp.ndarray]]:
        """Perform one step of the synapses and obtain their currents."""
        states = self._step_synapse_state(states, syn_channels, params, delta_t, edges)
        states, current_terms = self._synapse_currents(
            states, syn_channels, params, delta_t, edges
        )
        return states, current_terms

    def _step_synapse_state(
        self,
        states: Dict,
        syn_channels: List,
        params: Dict,
        delta_t: float,
        edges: pd.DataFrame,
    ) -> Dict:
        voltages = states["v"]

        grouped_syns = edges.groupby("type", sort=False, group_keys=False)
        pre_syn_inds = grouped_syns["pre_global_comp_index"].apply(list)
        post_syn_inds = grouped_syns["post_global_comp_index"].apply(list)
        synapse_names = list(grouped_syns.indices.keys())

        for i, synapse_type in enumerate(syn_channels):
            assert (
                synapse_names[i] == synapse_type._name
            ), "Mixup in the ordering of synapses. Please create an issue on Github."
            synapse_param_names = list(synapse_type.synapse_params.keys())
            synapse_state_names = list(synapse_type.synapse_states.keys())

            synapse_params = {}
            for p in synapse_param_names:
                synapse_params[p] = params[p]
            synapse_states = {}
            for s in synapse_state_names:
                synapse_states[s] = states[s]

            pre_inds = np.asarray(pre_syn_inds[synapse_names[i]])
            post_inds = np.asarray(post_syn_inds[synapse_names[i]])

            # State updates.
            states_updated = synapse_type.update_states(
                synapse_states,
                delta_t,
                voltages[pre_inds],
                voltages[post_inds],
                synapse_params,
            )

            # Rebuild state.
            for key, val in states_updated.items():
                states[key] = val

        return states

    def _synapse_currents(
        self,
        states: Dict,
        syn_channels: List,
        params: Dict,
        delta_t: float,
        edges: pd.DataFrame,
    ) -> Tuple[Dict, Tuple[jnp.ndarray, jnp.ndarray]]:
        voltages = states["v"]

        grouped_syns = edges.groupby("type", sort=False, group_keys=False)
        pre_syn_inds = grouped_syns["pre_global_comp_index"].apply(list)
        post_syn_inds = grouped_syns["post_global_comp_index"].apply(list)
        synapse_names = list(grouped_syns.indices.keys())

        syn_voltage_terms = jnp.zeros_like(voltages)
        syn_constant_terms = jnp.zeros_like(voltages)
        # Run with two different voltages that are `diff` apart to infer the slope and
        # offset.
        diff = 1e-3
        for i, synapse_type in enumerate(syn_channels):
            assert (
                synapse_names[i] == synapse_type._name
            ), "Mixup in the ordering of synapses. Please create an issue on Github."
            synapse_param_names = list(synapse_type.synapse_params.keys())
            synapse_state_names = list(synapse_type.synapse_states.keys())

            synapse_params = {}
            for p in synapse_param_names:
                synapse_params[p] = params[p]
            synapse_states = {}
            for s in synapse_state_names:
                synapse_states[s] = states[s]

            # Get pre and post indexes of the current synapse type.
            pre_inds = np.asarray(pre_syn_inds[synapse_names[i]])
            post_inds = np.asarray(post_syn_inds[synapse_names[i]])

            # Compute slope and offset of the current through every synapse.
            pre_v_and_perturbed = jnp.stack(
                [voltages[pre_inds], voltages[pre_inds] + diff]
            )
            post_v_and_perturbed = jnp.stack(
                [voltages[post_inds], voltages[post_inds] + diff]
            )
            synapse_currents = vmap(
                synapse_type.compute_current, in_axes=(None, 0, 0, None)
            )(
                synapse_states,
                pre_v_and_perturbed,
                post_v_and_perturbed,
                synapse_params,
            )
            synapse_currents_dist = convert_point_process_to_distributed(
                synapse_currents,
                params["radius"][post_inds],
                params["length"][post_inds],
            )

            # Split into voltage and constant terms.
            voltage_term = (synapse_currents_dist[1] - synapse_currents_dist[0]) / diff
            constant_term = (
                synapse_currents_dist[0] - voltage_term * voltages[post_inds]
            )

            # Gather slope and offset for every postsynaptic compartment.
            gathered_syn_currents = gather_synapes(
                len(voltages),
                post_inds,
                voltage_term,
                constant_term,
            )
            syn_voltage_terms += gathered_syn_currents[0]
            syn_constant_terms -= gathered_syn_currents[1]

            # Add the synaptic currents through every compartment as state.
            # `post_syn_currents` is a `jnp.ndarray` of as many elements as there are
            # compartments in the network.
            # `[0]` because we only use the non-perturbed voltage.
            states[f"i_{synapse_type._name}"] = synapse_currents[0]

        return states, (syn_voltage_terms, syn_constant_terms)

    def arrange_in_layers(
        self,
        layers: List[int],
        within_layer_offset: float = 500.0,
        between_layer_offset: float = 1500.0,
        vertical_layers: bool = False,
    ):
        """Arrange the cells in the network to form layers.

        Moves the cells in the network to arrange them into layers.

        Args:
            layers: List of integers specifying the number of cells in each layer.
            within_layer_offset: Offset between cells within the same layer.
            between_layer_offset: Offset between layers.
            vertical_layers: If True, layers are arranged vertically.
        """
        assert (
            np.sum(layers) == self.shape[0]
        ), "The number of cells in the layers must match the number of cells in the network."
        cells_in_layers = [
            list(range(sum(layers[:i]), sum(layers[: i + 1])))
            for i in range(len(layers))
        ]

        for l, cell_inds in enumerate(cells_in_layers):
            layer = self.cell(cell_inds)
            for i, cell in enumerate(layer.cells):
                if vertical_layers:
                    x_offset = (i - (len(cell_inds) - 1) / 2) * within_layer_offset
                    y_offset = (len(layers) - 1 - l) * between_layer_offset
                else:
                    x_offset = l * between_layer_offset
                    y_offset = (i - (len(cell_inds) - 1) / 2) * within_layer_offset

                cell.move_to(x=x_offset, y=y_offset, z=0)

    def vis(
        self,
        detail: str = "full",
        ax: Optional[Axes] = None,
        color: str = "k",
        synapse_color: str = "b",
        dims: Tuple[int] = (0, 1),
        cell_plot_kwargs: Dict = {},
        synapse_plot_kwargs: Dict = {},
        synapse_scatter_kwargs: Dict = {},
        **kwargs,  # absorb add. kwargs, i.e. to enable net.cell(0).vis(type="line")
    ) -> Axes:
        """Visualize the module.

        Args:
            detail: Either of [point, full]. `point` visualizes every neuron in the
                network as a dot.
                `full` plots the full morphology of every neuron. It requires that
                `compute_xyz()` has been run.
            color: The color in which cells are plotted.
            synapse_color: The color in which synapses are plotted.
            dims: Which dimensions to plot. 1=x, 2=y, 3=z coordinate. Must be a tuple of
                two of them.
            cell_plot_kwargs: Keyword arguments passed to the plotting function for
                cell morphologies. Only takes effect for `detail='full'`.
            synapse_plot_kwargs: Keyword arguments passed to the plotting function for
                syanpses.
            synapse_scatter_kwargs: Keyword arguments passed to the scatter function for
                syanpse terminals.
        """
        xyz0 = self.cell(0).xyzr[0][:, :3]
        same_xyz = np.all([np.all(xyz0 == cell.xyzr[0][:, :3]) for cell in self.cells])
        if same_xyz:
            warn(
                "Same coordinates for all cells. Consider using `move`, `move_to` or `arrange_in_layers` to move them."
            )

        if ax is None:
            fig = plt.figure(figsize=(3, 3))
            ax = fig.add_subplot(111) if len(dims) < 3 else plt.axes(projection="3d")

        # detail="point" -> pos taken to be the mean of all traced points on the cell.
        cell_to_point_xyz = lambda cell: np.mean(np.vstack(cell.xyzr)[:, :3], axis=0)

        dims_np = np.asarray(dims)
        if detail == "point":
            for cell in self.cells:
                pos = cell_to_point_xyz(cell)[dims_np]
                ax.scatter(*pos, color=color, **cell_plot_kwargs)
        elif detail == "full":
            ax = super().vis(dims=dims, color=color, ax=ax, **cell_plot_kwargs)
        else:
            raise ValueError("detail must be in {full, point}.")

        nodes = self.nodes.set_index("global_comp_index")
        for i, edge in self.edges.iterrows():
            prepost_locs = []
            for prepost in ["pre", "post"]:
                loc, comp = edge[[prepost + "_locs", prepost + "_global_comp_index"]]
                branch = nodes.loc[comp, "global_branch_index"]
                cell = nodes.loc[comp, "global_cell_index"]
                branch_xyz = self.xyzr[branch][:, :3]

                xyz_loc = branch_xyz
                if detail == "point":
                    xyz_loc = cell_to_point_xyz(self.cell(cell))
                elif len(branch_xyz) == 2:
                    # If only start and end point of a branch are traced, perform a
                    # linear interpolation to get the synpase location.
                    xyz_loc = branch_xyz[0] + (branch_xyz[1] - branch_xyz[0]) * loc
                else:
                    # If densely traced, use intermediate trace values for synapse loc.
                    middle_ind = int((len(branch_xyz) - 1) * loc)
                    xyz_loc = xyz_loc[middle_ind]

                prepost_locs.append(xyz_loc)
            prepost_locs = np.stack(prepost_locs).T
            ax.plot(*prepost_locs[dims_np], color=synapse_color, **synapse_plot_kwargs)
            ax.scatter(
                *prepost_locs[dims_np, 1], color=synapse_color, **synapse_scatter_kwargs
            )

        return ax

    def _infer_synapse_type_ind(self, synapse_name):
        syn_names = self.base.synapse_names
        is_new_type = False if synapse_name in syn_names else True
        type_ind = len(syn_names) if is_new_type else syn_names.index(synapse_name)
        return type_ind, is_new_type

    def _update_synapse_state_names(self, synapse_type):
        # (Potentially) update variables that track meta information about synapses.
        self.base.synapse_names.append(synapse_type._name)
        self.base.synapse_param_names += list(synapse_type.synapse_params.keys())
        self.base.synapse_state_names += list(synapse_type.synapse_states.keys())
        self.base.synapses.append(synapse_type)

    def _append_multiple_synapses(self, pre_nodes, post_nodes, synapse_type):
        # Add synapse types to the module and infer their unique identifier.
        synapse_name = synapse_type._name
        synapse_current_name = f"i_{synapse_name}"
        type_ind, is_new = self._infer_synapse_type_ind(synapse_name)
        if is_new:  # synapse is not known
            self._update_synapse_state_names(synapse_type)
            self.base.synapse_current_names.append(synapse_current_name)

        index = len(self.base.edges)
        indices = [idx for idx in range(index, index + len(pre_nodes))]
        global_edge_index = pd.DataFrame({"global_edge_index": indices})
        post_loc = loc_of_index(
            post_nodes["global_comp_index"].to_numpy(),
            post_nodes["global_branch_index"].to_numpy(),
            self.ncomp_per_branch,
        )
        pre_loc = loc_of_index(
            pre_nodes["global_comp_index"].to_numpy(),
            pre_nodes["global_branch_index"].to_numpy(),
            self.ncomp_per_branch,
        )

        # Define new synapses. Each row is one synapse.
        pre_nodes = pre_nodes[["global_comp_index"]]
        pre_nodes.columns = ["pre_global_comp_index"]
        post_nodes = post_nodes[["global_comp_index"]]
        post_nodes.columns = ["post_global_comp_index"]
        new_rows = pd.concat(
            [
                global_edge_index,
                pre_nodes.reset_index(drop=True),
                post_nodes.reset_index(drop=True),
            ],
            axis=1,
        )
        new_rows["type"] = synapse_name
        new_rows["type_ind"] = type_ind
        new_rows["pre_locs"] = pre_loc
        new_rows["post_locs"] = post_loc
        self.base.edges = concat_and_ignore_empty(
            [self.base.edges, new_rows], ignore_index=True, axis=0
        )
        self._add_params_to_edges(synapse_type, indices)
        self.base.edges["controlled_by_param"] = 0
        self._edges_in_view = self.edges.index.to_numpy()

    def _add_params_to_edges(self, synapse_type, indices):
        # Add parameters and states to the `.edges` table.
        for key, param_val in synapse_type.synapse_params.items():
            self.base.edges.loc[indices, key] = param_val

        # Update synaptic state array.
        for key, state_val in synapse_type.synapse_states.items():
            self.base.edges.loc[indices, key] = state_val



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

from typing import List

import numpy as np

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.synapses import IonotropicSynapse, Synapse, TestSynapse


def test_set_and_querying_params_one_type(SimpleNet):
    """Test if the correct parameters are set if one type of synapses is inserted."""
    net = SimpleNet(4, 1, 4)

    for pre_ind in [0, 1]:
        for post_ind in [2, 3]:
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, IonotropicSynapse())

    # Get the synapse parameters to test setting
    syn_params = list(IonotropicSynapse().synapse_params.keys())
    for p in syn_params:
        net.set(p, 0.15)
        assert np.all(net.edges[p].to_numpy() == 0.15)

    full_syn_view = net.IonotropicSynapse
    single_syn_view = net.IonotropicSynapse.edge(1)
    double_syn_view = net.IonotropicSynapse.edge([2, 3])

    # There shouldn't be too many synapse_params otherwise this will take a long time
    for p in syn_params:
        full_syn_view.set(p, 0.32)
        assert np.all(net.edges[p].to_numpy() == 0.32)

        single_syn_view.set(p, 0.18)
        assert net.edges[p].to_numpy()[1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([0, 2, 3])] == 0.32)

        double_syn_view.set(p, 0.12)
        assert net.edges[p][0] == 0.32
        assert net.edges[p][1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([2, 3])] == 0.12)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional

import jax.numpy as jnp

from jaxley.pumps.pump import Pump


def convert_current_to_molarity_change(
    current: float, valence: float, radius: float
) -> float:
    r"""Compute the change in molarity of an ion given current returned by a channel.

    This function returns the _change_ in molarity M, i.e. :math:`\frac{dM}{dt}`.
    Consistent with the units of Jaxley (mM for concentration, ms for time), it returns
    the unit milli-mole / liter / milli-second.

    This function implements the formula

    .. math::

        dM/dt = \frac{I \cdot A}{V \cdot F \cdot z},

    where:

    * \( I \) is the current per surface area,
    * \( A \) is the surface area of a compartment,
    * \( V \) is the volume of a compartment,
    * \( F \) is the Faraday constant,
    * \( z \) is the valence of the ion.

    Args:
        current: Current through an ion channel (returned by the `compute_current`
            method of a `Channel`) in mS/cm^2.
        valence: Valence of the ion (e.g., 2 for calcium, 1 for sodium).
        radius: Radius of the compartment in micrometer.

    Returns:
        The change in molarity M, dM/dt, in unit milli-mole/liter/milli-second.
    """
    FARADAY = 96485.3329  # Coulombs per mole.

    # surface_area = 2 * pi * radius * length  # um^2
    # volume = pi * radius ** 2 * length  # um^3
    # surface_per_volume = surface_area / volume = 2 / radius
    surface_per_volume = 2 / radius

    # Where does 1e4 come from? `volume`: current * area / volume ->
    # mA / cm^2 * um^2 / um^3 = mA / cm^2 / um = mA / dm^2 * (1e1)^2 / dm * 1e5 =
    # mA / dm^3 * 1e7 = mA / liter * 1e7
    #
    # Finally, FARADAY is A * second / mol. However, our time is in units ms (not s),
    # so we need to divide by 1000 -> 1e7 / 1e3 = 1e4
    return current * surface_per_volume * 1e4 / FARADAY / valence


class CaFaradayConcentrationChange(Pump):
    r"""Update the intracellular calcium ion concentration depending on calcium current.

    This channel implements Faraday's first law of electrolysis to update the
    intracellular calcium concentration based on calcium current. Faraday's law relates
    how a current (e.g., through a channel) impacts the number of ions. Mathematically:

    .. math::

        n = \frac{I \cdot t}{F \cdot z}

    Taking the derivative with respect to time:

    .. math::

        \frac{dn}{dt} = \frac{I}{F \cdot z}

    where:

    * \( n \) is the amount of substance (number of moles),
    * \( I \) is the current,
    * \( t \) is time,
    * \( F \) is the Faraday constant,
    * \( z \) is the valence of the ion.

    To obtain concentration \( c \) from the amount of substance \( n \), we divide by
    the volume:

    .. math::

        \frac{dc}{dt} = \frac{1}{V} \cdot \frac{dn}{dt} = \frac{I}{F \cdot z \cdot V}

    In Jaxley, the current is given in mS/cm2, so we first have to multiply the current
    by the surface area of a compartment.

    The update is fully passive (i.e., there is no active pump). As such, it is even
    possible that ion concentration can become negative (because we do not enforce
    that calcium currents stop when no more ions are available).
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        self.channel_params = {}
        self.channel_states = {"i_Ca": 1e-8, "CaCon_i": 5e-05}
        self.ion_name = "CaCon_i"
        self.current_name = "i_CaCurrent"
        self.META = {"mechanism": "Calcium change"}

    def update_states(
        self,
        states: Dict[str, jnp.ndarray],
        dt,
        v,
        params: Dict[str, jnp.ndarray],
    ):
        """Update states if necessary (but this mechanism has no states to update)."""
        return {"CaCon_i": states["CaCon_i"], "i_Ca": states["i_Ca"]}

    def compute_current(
        self,
        states: Dict[str, jnp.ndarray],
        modified_state,
        params: Dict[str, jnp.ndarray],
    ):
        """Return change of calcium concentration as the calcium current."""
        # 2.0 is valence of calcium.
        return convert_current_to_molarity_change(states["i_Ca"], 2.0, params["radius"])

    def init_state(
        self,
        states: Dict[str, jnp.ndarray],
        v: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp
import numpy as np
import pandas as pd
from matplotlib.axes import Axes

from jaxley.modules.base import Module
from jaxley.utils.cell_utils import compute_children_and_parents
from jaxley.utils.misc_utils import cumsum_leading_zero
from jaxley.utils.solver_utils import JaxleySolveIndexer, comp_edges_to_indices


class Compartment(Module):
    """Compartment class.

    This class defines a single compartment that can be simulated by itself or
    connected up into branches. It is the basic building block of a neuron model.
    """

    compartment_params: Dict = {
        "length": 10.0,  # um
        "radius": 1.0,  # um
        "axial_resistivity": 5_000.0,  # ohm cm
        "capacitance": 1.0,  # uF/cm^2
    }
    compartment_states: Dict = {"v": -70.0}

    def __init__(self):
        super().__init__()

        self.ncomp = 1
        self.ncomp_per_branch = np.asarray([1])
        self.total_nbranches = 1
        self.nbranches_per_cell = [1]
        self._cumsum_nbranches = np.asarray([0, 1])
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)

        # Setting up the `nodes` for indexing.
        self.nodes = pd.DataFrame(
            dict(global_cell_index=[0], global_branch_index=[0], global_comp_index=[0])
        )
        self._append_params_and_states(self.compartment_params, self.compartment_states)
        self._update_local_indices()
        self._init_view()

        # Synapses.
        self.branch_edges = pd.DataFrame(
            dict(parent_branch_index=[], child_branch_index=[])
        )

        # For morphology indexing.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )
        self._internal_node_inds = jnp.asarray([0])

        # Initialize the module.
        self._initialize()

        # Coordinates.
        self.xyzr = [float("NaN") * np.zeros((2, 4))]

    def _init_morph_jaxley_spsolve(self):
        self._solve_indexer = JaxleySolveIndexer(
            cumsum_ncomp=self.cumsum_ncomp,
            ncomp_per_branch=self.ncomp_per_branch,
            branchpoint_group_inds=np.asarray([]).astype(int),
            children_in_level=[],
            parents_in_level=[],
            root_inds=np.asarray([0]),
            remapped_node_indices=self._internal_node_inds,
        )

    def _init_morph_jax_spsolve(self):
        """Initialize morphology for the jax sparse voltage solver.

        Explanation of `self._comp_eges['type']`:
        `type == 0`: compartment <--> compartment (within branch)
        `type == 1`: branchpoint --> parent-compartment
        `type == 2`: branchpoint --> child-compartment
        `type == 3`: parent-compartment --> branchpoint
        `type == 4`: child-compartment --> branchpoint
        """
        self._comp_edges = pd.DataFrame().from_dict(
            {"source": [], "sink": [], "type": []}
        )
        n_nodes, data_inds, indices, indptr = comp_edges_to_indices(self._comp_edges)
        self._n_nodes = n_nodes
        self._data_inds = data_inds
        self._indices_jax_spsolve = indices
        self._indptr_jax_spsolve = indptr



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from functools import partial
from typing import Callable, List, Optional, Tuple
from warnings import warn

import jax.numpy as jnp
import numpy as np

from jaxley.io.graph import from_graph, swc_to_graph
from jaxley.modules import Branch, Cell, Compartment
from jaxley.utils.cell_utils import build_radiuses_from_xyzr
from jaxley.utils.misc_utils import deprecated_kwargs


def _split_long_branches(
    branches: np.ndarray,
    types: np.ndarray,
    content: np.ndarray,
    max_branch_len: float,
    is_single_point_soma: bool,
) -> Tuple[np.ndarray, np.ndarray]:
    pathlengths = _compute_pathlengths(
        branches, content[:, 1:6], is_single_point_soma=is_single_point_soma
    )
    pathlengths = [np.sum(length_traced) for length_traced in pathlengths]
    split_branches = []
    split_types = []
    for branch, type, length in zip(branches, types, pathlengths):
        num_subbranches = 1
        split_branch = [branch]
        while length > max_branch_len:
            num_subbranches += 1
            split_branch = _split_branch_equally(branch, num_subbranches)
            lengths_of_subbranches = _compute_pathlengths(
                split_branch,
                coords=content[:, 1:6],
                is_single_point_soma=is_single_point_soma,
            )
            lengths_of_subbranches = [
                np.sum(length_traced) for length_traced in lengths_of_subbranches
            ]
            length = max(lengths_of_subbranches)
            if num_subbranches > 10:
                warn(
                    """`num_subbranches > 10`, stopping to split. Most likely your
                     SWC reconstruction is not dense and some neighbouring traced
                     points are farther than `max_branch_len` apart."""
                )
                break
        split_branches += split_branch
        split_types += [type] * num_subbranches

    return split_branches, split_types


def _split_branch_equally(branch: np.ndarray, num_subbranches: int) -> List[np.ndarray]:
    num_points_each = len(branch) // num_subbranches
    branches = [branch[:num_points_each]]
    for i in range(1, num_subbranches - 1):
        branches.append(branch[i * num_points_each - 1 : (i + 1) * num_points_each])
    branches.append(branch[(num_subbranches - 1) * num_points_each - 1 :])
    return branches


def _split_into_branches(
    content: np.ndarray, is_single_point_soma: bool
) -> Tuple[np.ndarray, np.ndarray]:
    """Separates the list of parents from SWC into a list of branch-lists.

    Each branch list contains the index of the first until last SWC node within the
    branch. E.g., the returned `all_branches` might look like:
    ```python
    [[0, 1, 2], [2, 3], [2, 4]]
    ```
    """
    prev_ind = None
    prev_type = None
    n_branches = 0

    # Branch inds will contain the row identifier at which a branch point occurs
    # (i.e. the row of the parent of two branches).
    branch_inds = []
    for c in content:
        current_ind = c[0]
        current_parent = c[-1]
        current_type = c[1]
        if current_parent != prev_ind or current_type != prev_type:
            branch_inds.append(int(current_parent))
            n_branches += 1
        prev_ind = current_ind
        prev_type = current_type

    all_branches = []
    current_branch = []
    all_types = []

    # `previous_type` tracks the type (soma, axon...) of the previous SWC node.
    previous_type = 0
    for c in content:  # Loop over every line in the SWC file.
        current_ind = int(c[0])  # First col is row_identifier
        current_parent = int(c[-1])  # Last col is parent in SWC specification.
        current_type = int(c[1])  # Second col is the type (soma, axon, basal,...).

        # Single point somas will be ignore below (`if len(current_branch) > 1`), so
        # we specifically have to add them here.
        if current_parent == -1 and is_single_point_soma and current_ind == 1:
            all_branches.append([current_ind])
            all_types.append(current_type)

        # Either append the current point to the branch, or add the branch to
        # `all_branches`.
        if current_parent in branch_inds[1:]:
            if len(current_branch) > 1:
                # Note that the `current_ind` is not part of `current_branch`! This is
                # because it already belongs to the next branch. This is also the
                # reason why we have to append the `previous_type` as type here: the
                # `current_type` will already be the type of the next branch, which can
                # be of a different type.
                all_branches.append(current_branch)
                all_types.append(previous_type)
            current_branch = [int(current_parent), int(current_ind)]
        else:
            current_branch.append(int(current_ind))

        previous_type = current_type

    # Append the final branch (intermediate branches are already appended five lines
    # above.)
    all_branches.append(current_branch)
    all_types.append(current_type)
    return all_branches, all_types


def _split_into_branches_and_sort(
    content: np.ndarray,
    max_branch_len: Optional[float],
    is_single_point_soma: bool,
    sort: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    branches, types = _split_into_branches(content, is_single_point_soma)
    if max_branch_len is not None:
        branches, types = _split_long_branches(
            branches,
            types,
            content,
            max_branch_len,
            is_single_point_soma=is_single_point_soma,
        )

    if sort:
        first_val = np.asarray([b[0] for b in branches])
        sorting = np.argsort(first_val, kind="mergesort")
        sorted_branches = [branches[s] for s in sorting]
        sorted_types = [types[s] for s in sorting]
    else:
        sorted_branches = branches
        sorted_types = types
    return sorted_branches, sorted_types


def _radius_generating_fns(
    all_branches: np.ndarray,
    radiuses: np.ndarray,
    each_length: np.ndarray,
    parents: np.ndarray,
    types: np.ndarray,
) -> List[Callable]:
    """For all branches in a cell, returns callable that return radius given loc."""
    radius_fns = []
    for i, branch in enumerate(all_branches):
        rads_in_branch = radiuses[np.asarray(branch) - 1]
        if parents[i] > -1 and types[i] != types[parents[i]]:
            # We do not want to linearly interpolate between the radius of the previous
            # branch if a new type of neurite is found (e.g. switch from soma to
            # apical). From looking at the SWC from n140.swc I believe that this is
            # also what NEURON does.
            rads_in_branch[0] = rads_in_branch[1]
        radius_fn = _radius_generating_fn(
            radiuses=rads_in_branch, each_length=each_length[i]
        )
        # Beause SWC starts counting at 1, but numpy counts from 0.
        # ind_of_branch_endpoint = np.asarray(b) - 1
        radius_fns.append(radius_fn)
    return radius_fns


def _padded_radius(loc: float, radiuses: np.ndarray) -> float:
    return radiuses * np.ones_like(loc)


def _radius(loc: float, cutoffs: np.ndarray, radiuses: np.ndarray) -> float:
    """Function which returns the radius via linear interpolation.

    Defined outside of `_radius_generating_fns` to allow for pickling of the resulting
    Cell object."""
    index = np.digitize(loc, cutoffs, right=False)
    left_rad = radiuses[index - 1]
    right_rad = radiuses[index]
    left_loc = cutoffs[index - 1]
    right_loc = cutoffs[index]
    loc_within_bin = (loc - left_loc) / (right_loc - left_loc)
    return left_rad + (right_rad - left_rad) * loc_within_bin


def _padded_radius_generating_fn(radiuses: np.ndarray) -> Callable:
    return partial(_padded_radius, radiuses=radiuses)


def _radius_generating_fn(radiuses: np.ndarray, each_length: np.ndarray) -> Callable:
    # Avoid division by 0 with the `summed_len` below.
    each_length[each_length < 1e-8] = 1e-8
    summed_len = np.sum(each_length)
    cutoffs = np.cumsum(np.concatenate([np.asarray([0]), each_length])) / summed_len
    cutoffs[0] -= 1e-8
    cutoffs[-1] += 1e-8

    # We have to linearly interpolate radiuses, therefore we need at least two radiuses.
    # However, jaxley allows somata which consist of a single traced point (i.e.
    # just one radius). Therefore, we just `tile` in order to generate an artificial
    # endpoint and startpoint radius of the soma.
    if len(radiuses) == 1:
        radiuses = np.tile(radiuses, 2)

    return partial(_radius, cutoffs=cutoffs, radiuses=radiuses)


def _build_parents(all_branches: List[np.ndarray]) -> List[int]:
    parents = [None] * len(all_branches)
    all_last_inds = [b[-1] for b in all_branches]
    for i, branch in enumerate(all_branches):
        parent_ind = branch[0]
        ind = np.where(np.asarray(all_last_inds) == parent_ind)[0]
        if len(ind) > 0 and ind != i:
            parents[i] = ind[0]
        else:
            assert (
                parent_ind == 1
            ), """Trying to connect a segment to the beginning of 
            another segment. This is not allowed. Please create an issue on github."""
            parents[i] = -1

    return parents


def _compute_pathlengths(
    all_branches: np.ndarray, coords: np.ndarray, is_single_point_soma: bool
) -> List[np.ndarray]:
    """
    Args:
        coords: Has shape (num_traced_points, 5), where `5` is (type, x, y, z, radius).
    """
    branch_pathlengths = []
    for b in all_branches:
        coords_in_branch = coords[np.asarray(b) - 1]
        if len(coords_in_branch) > 1:
            # If the branch starts at a different neurite (e.g. the soma) then NEURON
            # ignores the distance from that initial point. To reproduce, use the
            # following SWC dummy file and read it in NEURON (and Jaxley):
            # 1 1 0.00 0.0 0.0 6.0 -1
            # 2 2 9.00 0.0 0.0 0.5 1
            # 3 2 10.0 0.0 0.0 0.3 2
            types = coords_in_branch[:, 0]
            if int(types[0]) == 1 and int(types[1]) != 1 and is_single_point_soma:
                coords_in_branch[0] = coords_in_branch[1]

            # Compute distances between all traced points in a branch.
            point_diffs = np.diff(coords_in_branch, axis=0)
            dists = np.sqrt(
                point_diffs[:, 1] ** 2 + point_diffs[:, 2] ** 2 + point_diffs[:, 3] ** 2
            )
        else:
            # Jaxley uses length and radius for every compartment and assumes the
            # surface area to be 2*pi*r*length. For branches consisting of a single
            # traced point we assume for them to have area 4*pi*r*r. Therefore, we have
            # to set length = 2*r.
            radius = coords_in_branch[0, 4]  # txyzr -> 4 is radius.
            dists = np.asarray([2 * radius])
        branch_pathlengths.append(dists)
    return branch_pathlengths


def swc_to_jaxley(
    fname: str,
    max_branch_len: Optional[float] = None,
    sort: bool = True,
    num_lines: Optional[int] = None,
) -> Tuple[List[int], List[float], List[Callable], List[float], List[np.ndarray]]:
    """Read an SWC file and bring morphology into `jaxley` compatible formats.

    Args:
        fname: Path to swc file.
        max_branch_len: Maximal length of one branch. If a branch exceeds this length,
            it is split into equal parts such that each subbranch is below
            `max_branch_len`.
        num_lines: Number of lines of the SWC file to read.
    """
    content = np.loadtxt(fname)[:num_lines]
    types = content[:, 1]
    is_single_point_soma = types[0] == 1 and types[1] != 1

    if is_single_point_soma:
        # Warn here, but the conversion of the length happens in `_compute_pathlengths`.
        warn(
            "Found a soma which consists of a single traced point. `Jaxley` "
            "interprets this soma as a spherical compartment with radius "
            "specified in the SWC file, i.e. with surface area 4*pi*r*r."
        )
    sorted_branches, types = _split_into_branches_and_sort(
        content,
        max_branch_len=max_branch_len,
        is_single_point_soma=is_single_point_soma,
        sort=sort,
    )

    parents = _build_parents(sorted_branches)
    each_length = _compute_pathlengths(
        sorted_branches, content[:, 1:6], is_single_point_soma=is_single_point_soma
    )
    pathlengths = [np.sum(length_traced) for length_traced in each_length]
    for i, pathlen in enumerate(pathlengths):
        if pathlen == 0.0:
            warn("Found a segment with length 0. Clipping it to 1.0")
            pathlengths[i] = 1.0
    radius_fns = _radius_generating_fns(
        sorted_branches, content[:, 5], each_length, parents, types
    )

    if np.sum(np.asarray(parents) == -1) > 1.0:
        parents = np.asarray([-1] + parents)
        parents[1:] += 1
        parents = parents.tolist()
        pathlengths = [0.1] + pathlengths
        radius_fns = [_padded_radius_generating_fn(content[0, 5])] + radius_fns
        sorted_branches = [[0]] + sorted_branches

        # Type of padded section is assumed to be of `custom` type:
        # http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
        types = [5.0] + types

    all_coords_of_branches = []
    for i, branch in enumerate(sorted_branches):
        # Remove 1 because `content` is an array that is indexed from 0.
        branch = np.asarray(branch) - 1

        # Deal with additional branch that might have been added above in the lines
        # `if np.sum(np.asarray(parents) == -1) > 1.0:`
        branch[branch < 0] = 0

        # Get traced coordinates of the branch.
        coords_of_branch = content[branch, 2:6]
        all_coords_of_branches.append(coords_of_branch)

    return parents, pathlengths, radius_fns, types, all_coords_of_branches


def read_swc_custom(
    fname: str,
    ncomp: Optional[int] = None,
    max_branch_len: Optional[float] = None,
    min_radius: Optional[float] = None,
    assign_groups: bool = True,
) -> Cell:
    """Reads SWC file into a `Cell`.

    Jaxley assumes cylindrical compartments and therefore defines length and radius
    for every compartment. The surface area is then 2*pi*r*length. For branches
    consisting of a single traced point we assume for them to have area 4*pi*r*r.
    Therefore, in these cases, we set lenght=2*r.

    Args:
        fname: Path to the swc file.
        ncomp: The number of compartments per branch.
        max_branch_len: If a branch is longer than this value it is split into two
            branches.
        min_radius: If the radius of a reconstruction is below this value it is clipped.
        assign_groups: If True, then the identity of reconstructed points in the SWC
            file will be used to generate groups `undefined`, `soma`, `axon`, `basal`,
            `apical`, `custom`. See here:
            http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html

    Returns:
        A `Cell` object.
    """
    parents, pathlengths, radius_fns, types, coords_of_branches = swc_to_jaxley(
        fname, max_branch_len=max_branch_len, sort=True, num_lines=None
    )
    nbranches = len(parents)

    comp = Compartment()
    branch = Branch([comp for _ in range(ncomp)])
    cell = Cell(
        [branch for _ in range(nbranches)], parents=parents, xyzr=coords_of_branches
    )
    # Also save the radius generating functions in case users post-hoc modify the number
    # of compartments with `.set_ncomp()`.
    cell._radius_generating_fns = radius_fns

    lengths_each = np.repeat(pathlengths, ncomp) / ncomp
    cell.set("length", lengths_each)

    radiuses_each = build_radiuses_from_xyzr(
        radius_fns,
        range(len(parents)),
        min_radius,
        ncomp,
    )
    cell.set("radius", radiuses_each)

    # Description of SWC file format:
    # http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
    ind_name_lookup = {
        0: "undefined",
        1: "soma",
        2: "axon",
        3: "basal",
        4: "apical",
        5: "custom",
    }
    types = np.asarray(types).astype(int)
    if assign_groups:
        for type_ind in np.unique(types):
            if type_ind < 5.5:
                name = ind_name_lookup[type_ind]
            else:
                name = f"custom{type_ind}"
            indices = np.where(types == type_ind)[0].tolist()
            if len(indices) > 0:
                cell.branch(indices).add_to_group(name)
    return cell


def read_swc(
    fname: str,
    ncomp: Optional[int] = None,
    max_branch_len: Optional[float] = None,
    min_radius: Optional[float] = None,
    assign_groups: bool = True,
    backend: str = "custom",
    **backend_kwargs,
) -> Cell:
    """Reads SWC file into a `Cell`.

    Jaxley assumes cylindrical compartments and therefore defines length and radius
    for every compartment. The surface area is then 2*pi*r*length. For branches
    consisting of a single traced point we assume for them to have area 4*pi*r*r.
    Therefore, in these cases, we set lenght=2*r.

    Args:
        fname: Path to the swc file.
        ncomp: The number of compartments per branch.
        max_branch_len: If a branch is longer than this value it is split into two
            branches.
        min_radius: If the radius of a reconstruction is below this value it is clipped.
        assign_groups: If True, then the identity of reconstructed points in the SWC
            file will be used to generate groups `undefined`, `soma`, `axon`, `basal`,
            `apical`, `custom`. See here:
            http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
        backend: The backend to use. Currently `custom` and `graph` are supported.
            For context on these backends see `read_swc_custom` and `from_graph`.
        backend_kwargs: Additional keyword arguments passed to the swc reader.

    Returns:
        A `Cell` object."""

    if backend == "custom":
        return read_swc_custom(
            fname,
            ncomp=ncomp,
            max_branch_len=max_branch_len,
            min_radius=min_radius,
            assign_groups=assign_groups,
            **backend_kwargs,
        )
    elif backend == "graph":
        graph = swc_to_graph(fname)
        return from_graph(
            graph,
            ncomp=ncomp,
            max_branch_len=max_branch_len,
            min_radius=min_radius,
            assign_groups=assign_groups,
            **backend_kwargs,
        )
    else:
        raise ValueError(f"Unknown backend: {backend}. Use either `custom` or `graph`.")



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

import jax.numpy as jnp
import networkx as nx
import numpy as np
import pandas as pd
from jax import vmap

from jaxley.modules import Branch, Cell, Compartment, Network
from jaxley.utils.cell_utils import v_interp

# helper functions
_is_leaf = lambda G, n: G.out_degree(n) == 0 and G.in_degree(n) == 1
_is_root = lambda G, n: G.in_degree(n) == 0
_is_branching = lambda G, n: G.out_degree(n) > 1
_has_same_id = lambda G, i, j: G.nodes[i]["id"] == G.nodes[j]["id"]
_get_soma_idxs = lambda G: [
    i for i, n in nx.get_node_attributes(G, "id").items() if n == 1
]
_unpack = lambda d, keys: [d[k] for k in keys]
_branch_e2n = lambda b: np.unique(np.concatenate(b)).tolist()
_branch_n2e = lambda b: [e for e in zip(b[:-1], b[1:])]


def _find_root(G):
    roots = [n for n in G.nodes if _is_root(G, n)]
    if len(roots) > 1:
        raise ValueError("Multiple roots found in graph.")
    return roots[0]


def swc_to_graph(fname: str, num_lines: int = None) -> nx.DiGraph:
    """Read a swc file and convert it to a networkx graph.

    The graph is read such that each entry in the swc file becomes a graph node
    with the column attributes (id, x, y, z, r). Then each node is connected to its
    designated parent via an edge. A "type" attribute is added to the graph to identify
    its processing stage for subsequent steps.

    Args:
        fname: Path to the swc file.
        num_lines: Number of lines to read from the file. If None, all lines are read.

    Returns:
        A networkx graph of the traced morphology in the swc file.
    """
    i_id_xyzr_p = np.loadtxt(fname)[:num_lines]

    graph = nx.DiGraph()
    graph.add_nodes_from(
        (
            (int(i), {"id": int(id), "x": x, "y": y, "z": z, "r": r})
            for i, id, x, y, z, r, p in i_id_xyzr_p
        )
    )
    graph.add_edges_from([(p, i) for p, i in i_id_xyzr_p[:, [-1, 0]] if p != -1])
    graph = nx.relabel_nodes(graph, {i: i - 1 for i in graph.nodes})
    return graph


def _get_nodes_and_parents(graph: nx.DiGraph) -> np.ndarray:
    """List (node, parent) pairs for a graph."""
    edges = []
    for node in graph.nodes():
        parents = list(graph.predecessors(node))
        edges += [(node, parent) for parent in parents] if parents else [(node, -1)]
    return np.array(edges)


def find_swc_trace_errors(graph: nx.DiGraph, ignore: Optional[List] = []) -> np.ndarray:
    """Identify discontinuities in the swc tracing order.

    Some swc files contain artefacts, where tracing of the same neurite was done
    in disconnected pieces. Both NEURON and Jaxley's hand coded swc reader introduce
    a break in the trace at these points, since they parse the file in order. This
    leads to split branches, which should be one. This function identifies these
    points in the graph.

    Example swc file:
    # branch 1
    1 1 0.0 0.0 0.0 1.0 -1
    2 1 1.0 0.0 0.0 1.0 1
    3 1 2.0 0.0 0.0 1.0 2
    # branch 2
    4 2 3.0 1.0 0.0 1.0 3
    5 2 4.0 2.0 0.0 1.0 4
    # branch 3
    6 3 3.0 -1.0 0.0 1.0 3
    7 3 4.0 -2.0 0.0 1.0 6
    8 3 5.0 -3.0 0.0 1.0 7
    # ammend branch 2
    9 4 5.0 3.0 0.0 1.0 5

    Args:
        graph: graph tracing of swc file (from `swc_to_graph`).
        ignore: Nodes to ignore.

    Returns:
        An array of node indices where tracing is discontinous.
    """
    edges = _get_nodes_and_parents(graph)[:, [-1, 0]]
    s, t = edges[0].T
    branch_ends = []

    # find first and last node of each connected segment, i.e.
    # [(1,2),(2,3),(3,4),(5,6),...] -> [(1,4),(5,...),...]
    for i, j in edges[1:]:
        if i != t:
            branch_ends += [(s, t)]
            s = i
        t = j
    branch_ends += [(s, t)]
    branch_ends = np.array(branch_ends)
    # nodes that end one and start another segment indicate a breakage during tracing
    break_idxs = branch_ends[:, 1][np.isin(branch_ends[:, 1], branch_ends[:, 0])]
    break_idxs = np.setdiff1d(break_idxs, ignore)
    return break_idxs


def simulate_swc_trace_errors(
    graph: nx.DiGraph, branches: List[np.ndarray], ignore: Optional[List] = []
) -> List[np.ndarray]:
    """Simulate swc trace errors in the branches.

    Both NEURON and Jaxley's hand coded swc reader introduce breaks in the trace
    if the same neurite was traced in disconnected pieces. Since `swc_to_graph` is
    agnostic to the order of the tracing, it does not suffer from this issue. Hence,
    to artifically force this behaviour (to compare to the other parsers), this
    function can be used to simulates these errors. See `find_swc_trace_errors` for
    how to identify these points in the graph.

    Args:
        graph: A networkx graph of a traced morphology.
        branches: List of branches represented as arrays of edges.
        ignore: Nodes to ignore when splitting branches.

    Returns:
        A list of branches with simulated trace errors.
    """
    node_idxs = find_swc_trace_errors(graph, ignore=ignore)

    for node_idx in node_idxs:
        branch_idx = next(i for i, p in enumerate(branches) if node_idx in p)
        b4, branch, after = (
            branches[:branch_idx],
            branches[branch_idx],
            branches[branch_idx + 1 :],
        )
        break_idx = np.where(branch == node_idx)[0][1]
        # insert artifical break into branch
        branches = b4 + [branch[:break_idx], branch[break_idx:]] + after
    return branches


def trace_branches(
    graph: nx.DiGraph, max_len=None, ignore_swc_trace_errors=True
) -> List[np.ndarray]:
    """Get all linearly connected paths in a graph aka. branches.

    The graph is traversed depth-first starting from the source node, which is the only
    node with in_degree 0 (raises in case multiple are found). Note: Traversal order can
    be changed, by reversing the edge directions, i.e. to start traversal from a leaf node
    set: [source -> ... -> leaf] to [leaf -> ... -> source].

    Args:
        graph: A networkx graph.
        ignore_swc_trace_errors: Whether to ignore discontinuities in the swc tracing
            order. If False, this will result in split branches at these points.
    Returns:
        A list of linear paths in the graph. Each path is represented as an array of
        edges."""

    # handles special case of a single soma node
    if len(soma_idxs := _get_soma_idxs(graph)) == 1:
        soma = soma_idxs[0]
        # edges connecting nodes to soma are considered part of the soma -> l = 0.
        for i, j in (*graph.in_edges(soma), *graph.out_edges(soma)):
            graph.edges[i, j]["l"] = 0

        # Setting l = 2*r ensures A_cylinder = 2*pi*r*l = 4*pi*r^2 = A_sphere
        graph.add_node(-1, **graph.nodes[0])
        graph.add_edge(-1, soma, l=2 * graph.nodes[soma]["r"])
        graph = nx.relabel_nodes(graph, {i: i + 1 for i in graph.nodes})

    # Ensure root segment is linear. Needed to create root branch.
    if graph.out_degree(0) > 1:
        # The root segment should be of type `custom` (=5).
        parent = graph.nodes[0]
        parent["id"] = 5
        graph.add_node(-1, **parent)
        graph.add_edge(-1, 0, l=0.1)
        graph = nx.relabel_nodes(graph, {i: i + 1 for i in graph.nodes})

    branches, current_branch = [], []

    root = _find_root(graph)
    for i, j in nx.dfs_edges(graph, root):
        current_branch += [(i, j)]
        if _is_leaf(graph, j) or _is_branching(graph, j):
            branches.append(current_branch)
            current_branch = []
        elif not _has_same_id(graph, i, j):  # start new branch if ids differ
            branches.append(current_branch[:-1])
            current_branch = [current_branch[-1]]

    branch_edges = [np.array(p) for p in branches if len(p) > 0]

    if max_len:
        edge_lens = nx.get_edge_attributes(graph, "l")
        branch_edges = split_branches(branch_edges, edge_lens, max_len)

    if not ignore_swc_trace_errors:
        # ignore added index by default; only relevant in case it was added
        branch_edges = simulate_swc_trace_errors(graph, branch_edges, ignore=[root])

    for br_idx, br_edges in enumerate(branch_edges):
        graph.add_edges_from(br_edges, branch_index=br_idx)
    return graph


def _add_edge_lens(graph: nx.DiGraph, min_len: float = 1e-5) -> nx.DiGraph:
    """Add edge lengths to graph.edges based on the xyz coordinates of graph.nodes."""
    xyz = lambda i: np.array(_unpack(graph.nodes[i], "xyz"))
    for i, j in graph.edges:
        d_ij = (
            np.sqrt(((xyz(i) - xyz(j)) ** 2).sum())
            if i != j
            else 2 * graph.nodes[i]["r"]
        )
        # min_len ensures that 2 nodes cannot lie on top of each other
        # this is important for the compartmentalization
        graph.edges[i, j]["l"] = d_ij if d_ij >= min_len else min_len
    return graph


def add_missing_graph_attrs(graph: nx.DiGraph) -> nx.DiGraph:
    """Add missing attributes to the graph nodes and edges.

    The following attributes are added to the graph:
    - id: int (default: 0)
    - x, y, z: float (default: NaN)
    - r: float (default: 1)
    - l: float (default: 1)

    Args:
        graph: A networkx graph.

    Returns:
        The graph with the added attributes."""
    available_keys = graph.nodes[0].keys()
    defaults = {
        "id": 0,
        "x": float("nan"),
        "y": float("nan"),
        "z": float("nan"),
        "r": 1,
    }
    # add defaults if not present
    for key in set(defaults.keys()).difference(available_keys):
        nx.set_node_attributes(graph, defaults[key], key)

    graph = _add_edge_lens(graph)
    edge_lens = nx.get_edge_attributes(graph, "l")
    if np.isnan(list(edge_lens.values())[0]):
        nx.set_edge_attributes(graph, 1, "l")

    return graph


def split_branches(
    branches: List[np.ndarray], edge_lens: Dict, max_len: int = 1000
) -> List[np.ndarray]:
    """Split branches into approximately equally long sections <= max_len.

    Args:
        branches: List of branches represented as arrays of edges.
        edge_lens: Dict for length of each edge in the graph.
        max_len: Maximum length of a branch section. If a branch exceeds this length,
            it is split into equal parts.

    Returns:
        A list of branches, where each branch is split into sections of length <= max_len.
    """
    # TODO: split branches into exactly equally long sections
    edge_lens.update({(j, i): l for (i, j), l in edge_lens.items()})
    new_branches = []
    for branch in branches:
        cum_branch_len = np.cumsum([edge_lens[i, j] for i, j in branch])
        k = cum_branch_len // max_len
        split_branch = [branch[np.where(np.array(k) == kk)[0]] for kk in np.unique(k)]
        new_branches += split_branch
    return new_branches


def insert_compartments(graph: nx.DiGraph, ncomp_per_branch: int) -> nx.DiGraph:
    """Insert compartment nodes into the graph.

    Inserts new nodes in every branch (edges with "branch_index" attribute) at equidistant
    points along it. Node attributes, like radius are linearly interpolated along its
    length.

    Example: 4 compartments | edges = - | nodes = o | comp_nodes = x
    o-----------o----------o---o---o---o--------o
    o-------x---o----x-----o--xo---o---ox-------o

    Args:
        graph: Mmorphology where edges are already labelled with "branch_index"
        ncomp_per_branch: How many compartments per branch to insert

    Returns:
        Graph with additional nodes that are labelled with "comp_index"
    """
    comp_offset = 0

    branch_inds = nx.get_edge_attributes(graph, "branch_index")
    branch_edge_df = pd.DataFrame(branch_inds.items(), columns=["edge", "branch_index"])
    for branch_index, branch_edges in branch_edge_df.groupby("branch_index")["edge"]:
        path_lens = np.cumsum([0] + [graph.edges[u, v]["l"] for u, v in branch_edges])
        branch_nodes = _branch_e2n(branch_edges.to_numpy())
        branch_data = pd.DataFrame([graph.nodes[i] for i in branch_nodes])
        branch_data["node_index"] = branch_nodes
        branch_data["l"] = path_lens

        # fix id and r bleed over from neighboring neurites of a different type
        if branch_data.loc[0, "id"] != branch_data.loc[1, "id"]:
            branch_data.loc[0, ["r", "id"]] = branch_data.loc[1, ["r", "id"]]

        branch_len = branch_data["l"].max()
        comp_len = branch_len / ncomp_per_branch
        locs = np.linspace(comp_len / 2, branch_len - comp_len / 2, ncomp_per_branch)

        new_branch_nodes = v_interp(locs, branch_data["l"].values, branch_data.values)
        new_branch_nodes = pd.DataFrame(
            np.array(new_branch_nodes.T), columns=branch_data.columns
        )
        new_branch_nodes["id"] = new_branch_nodes["id"].astype(int)
        new_branch_nodes["comp_length"] = comp_len
        new_branch_nodes["branch_index"] = branch_index
        new_branch_nodes["comp_index"] = comp_offset + np.arange(ncomp_per_branch)
        new_branch_nodes["node_index"] = (
            max(graph.nodes) + 1 + np.arange(ncomp_per_branch)
        )
        comp_offset += ncomp_per_branch

        # splice comps into morphology
        branch_data = pd.concat([branch_data, new_branch_nodes]).sort_values(
            "l", ignore_index=True
        )
        new_branch_nodes = new_branch_nodes.set_index("node_index")
        graph.add_nodes_from(new_branch_nodes.to_dict(orient="index").items())

        graph.remove_edges_from(branch_edges.to_numpy())
        new_branch_edges = _branch_n2e(branch_data["node_index"])
        graph.add_edges_from(new_branch_edges, branch_index=branch_index)

    # add missing edge lengths
    graph = _add_edge_lens(graph)

    # re-enumerate in dfs from root
    root = _find_root(graph)
    mapping = {old: new for new, old in enumerate(nx.dfs_preorder_nodes(graph, root))}
    graph = nx.relabel_nodes(graph, mapping)
    return graph


def _get_comp_edges_dfs(
    graph: nx.DiGraph, node: int, parent_comp: int = None, visited: set = None
) -> List[Tuple[int]]:
    """List edges between compartment nodes, ignoring non-compartment nodes.

    Traverses a graph depth first and only records nodes and their successors if they
    have a "compartment_index".

    Args:
        graph: Morphology with inserted compartment nodes.
        node: node_index from which to start depth first traversal
        parent_comp: node_index of parent compartment
        visited: Keeps track of visited nodes during traversal

    Returns:
        List of edges (parent_node_index, child_node_index) that directly connect
        compartments, while skipping indermediate hops via nodes w.o. "comp_index" attr.
    """
    if visited is None:
        visited = set()
    edges = []
    visited.add(node)

    current_comp = node if "comp_index" in graph.nodes[node] else None

    # If we have a parent with comp_index and current node has comp_index,
    # add the edge between them
    if parent_comp is not None and current_comp is not None:
        edges.append((parent_comp, current_comp))

    # Process all children depth first
    for child in graph.successors(node):
        if child not in visited:
            # Pass current_comp as parent if it exists, otherwise pass through the parent_comp
            prev_node = current_comp if current_comp is not None else parent_comp
            child_edges = _get_comp_edges_dfs(graph, child, prev_node, visited)
            edges.extend(child_edges)

    return edges


def extract_comp_graph(graph: nx.DiGraph) -> nx.DiGraph:
    """Get subgraph that only includes compartment nodes and their direct edges.

    Example: 4 compartments | edges = - | nodes = o | comp_nodes = x
    o-------x---o----x-----o--xo---o---ox-------o
            x--------x--------x---------x

    Args:
        graph: Morphology with compartment nodes

    Returns:
        Morphology with branches and compartments.
    """
    # create subgraph w.o. edges
    comp_nodes = list(nx.get_node_attributes(graph, "comp_index"))
    comp_graph = nx.subgraph(graph, comp_nodes).copy()

    # remove all edges
    comp_graph.remove_edges_from(list(comp_graph.edges))

    # find all edges between compartments and connect comp_graph
    root = _find_root(graph)
    comp_edges = _get_comp_edges_dfs(graph, root)
    comp_graph.add_edges_from(comp_edges)
    comp_graph = _add_edge_lens(comp_graph)

    for i, j in comp_graph.edges:
        branch_idx_i = graph.nodes[i]["branch_index"]
        branch_idx_j = graph.nodes[j]["branch_index"]
        if branch_idx_i != branch_idx_j:
            comp_graph.edges[i, j]["type"] = "inter_branch"
        else:
            comp_graph.edges[i, j]["type"] = "intra_branch"

    # fix 0 length edges (usually due adding root branch)
    min_len = 1e-3
    for i, j in comp_graph.edges:
        if comp_graph.edges[i, j]["l"] < min_len:
            comp_graph.edges[i, j]["l"] = min_len

    # relabel nodes using comp_index
    mapping = {n: attrs["comp_index"] for n, attrs in comp_graph.nodes(data=True)}

    comp_graph = nx.relabel_nodes(comp_graph, mapping, copy=True)
    return comp_graph


def make_jaxley_compatible(
    graph: nx.DiGraph,
    ncomp: int = 4,
    max_branch_len: float = None,
    min_radius: Optional[float] = None,
    ignore_swc_trace_errors: bool = True,
) -> nx.DiGraph:
    """Make a swc traced graph compatible with jaxley.

    In order to simulate a morphology jaxley imposes that it has to consist of
    branches, each of which consists of compartments with equal length. This function
    first traces the graph to compute the length of each edge and to identify its branch
    structure. It then splits these branches into equally long compartments. The
    compartment centers and radii are obtained by linearly interpolating the traced xyz
    coordinates and radii alongeach branch.

    This means initially, each node represents a traced point (an entry in the swc
    file) and each edge represents the direction along which the neurites were traced.
    Edges along every non-branching linear section of the neurite (smaller than the
    max_branch_len) are then assigned a branch index. Finally, each branch is split
    into pieces of equal length. This results in a new graph, where each node
    represents a compartment (at its center) and each edge represents a connection
    between these compartments. The graph is then ready to be imported into jaxley.

    Attributes that are added to the returned graph:
    - xyzr: list of the originally traced points.
    - type: the type of module is inferred depending on the number of compartments,
        i.e. if there is only one compartment, it will be "compartment". This let's
        `from_graph` know what type of module to return.
    - groups: the group of each compartment is inferred from the id attribute of the
        nodes.
    - cell_index: enumerates cells globally.
    - branch_index: enumerates branches globally.
    - comp_index: enumerates compartments globally.
    - radius: the radius of each compartment.
    - length: the length of each compartment.
    - x, y, z: the coordinates of each compartment.

    Args:
        graph: A networkx graph of a traced morphology.
        ncomp: The number of segments per compartment.
        max_branch_len: Maximal length of one branch. If a branch exceeds this length,
            it is split into equal parts.
        min_radius: If the radius of a reconstruction is below this value it is clipped.
        ignore_swc_trace_errors: Whether to ignore discontinuities in the swc tracing
            order. If False, this will result in split branches at these points.

    Returns:
        A networkx graph of the traced morphology that is compatible with jaxley
        and can be imported using `from_graph`. This is also the same format that
        is returned by `to_graph` when exporting a module.
    """
    # TODO: assert DAG and connected

    # pre-processing
    graph = add_missing_graph_attrs(graph)

    graph = trace_branches(graph, max_branch_len, ignore_swc_trace_errors)

    graph = insert_compartments(graph, ncomp)

    comp_graph = extract_comp_graph(graph)

    # post-processing

    # rename attrs
    # Description of SWC file format:
    # http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
    group_ids = {0: "undefined", 1: "soma", 2: "axon", 3: "basal", 4: "apical"}
    min_radius = min_radius if min_radius else 0.0
    clip_radius = lambda r: max(r, min_radius) if min_radius else r
    for n in comp_graph.nodes:
        # `dictionary.get(key, "custom")` returns `custom` by default. This is useful
        # because some SWC files might have really strange type identifier.
        type_id = comp_graph.nodes[n].pop("id")
        comp_graph.nodes[n]["groups"] = [group_ids.get(type_id, f"custom{type_id}")]
        comp_graph.nodes[n]["radius"] = clip_radius(comp_graph.nodes[n].pop("r"))
        comp_graph.nodes[n]["length"] = comp_graph.nodes[n].pop("comp_length")
        comp_graph.nodes[n].pop("l")

    nx.set_node_attributes(comp_graph, 0, "cell_index")

    # compute additional attributes
    edge_df = pd.DataFrame(
        [
            d["branch_index"] for i, j, d in graph.edges(data=True)
        ],  # TODO: filter out added comp indices to get original morphology
        index=graph.edges,
        columns=["branch_index"],
    )

    # xyzr
    edges_in_branches = edge_df.groupby("branch_index")
    nodes_in_branches = edges_in_branches.apply(
        lambda x: [
            n for n in _branch_e2n(x.index.values) if "comp_index" not in graph.nodes[n]
        ]
    )
    stack_branch_xyzr = lambda x: np.stack([_unpack(graph.nodes[n], "xyzr") for n in x])
    xyzr = nodes_in_branches.apply(stack_branch_xyzr).to_list()
    same_rows = lambda x: np.all(np.nan_to_num(x[0]) == np.nan_to_num(x))
    xyzr = [xyzr_i[[0, -1]] if same_rows(xyzr_i) else xyzr_i for xyzr_i in xyzr]

    comp_graph.graph["xyzr"] = xyzr
    comp_graph.graph["type"] = "cell"

    return comp_graph


def _infer_module_type_from_inds(idxs: pd.DataFrame) -> str:
    nuniques = idxs[["cell_index", "branch_index", "comp_index"]].nunique()
    nuniques.index = ["cell", "branch", "compartment"]
    nuniques = pd.concat([pd.Series({"network": 1}), nuniques])
    return_type = nuniques.loc[nuniques == 1].index[-1]
    return return_type


def build_module_scaffold(
    idxs: pd.DataFrame,
    return_type: Optional[str] = None,
    parent_branches: Optional[List[np.ndarray]] = None,
) -> Union[Network, Cell, Branch, Compartment]:
    """Builds a skeleton module from a DataFrame of indices.

    This is useful for instantiating a module that can be filled with data later.

    Args:
        idxs: DataFrame containing cell_index, branch_index, comp_index, i.e.
            Module.nodes or View.view.
        return_type: Type of module to return. If None, the type is inferred from the
            number of unique values in the indices. I.e. only 1 unique cell_index
                and 1 unique branch_index -> return_type = "jx.Branch".

    Returns:
        A skeleton module with the correct number of compartments, branches, cells, or
        networks."""
    return_types = ["compartment", "branch", "cell", "network"]
    build_cache = {k: [] for k in return_types}

    if return_type is None:  # infer return type from idxs
        return_type = _infer_module_type_from_inds(idxs)

    comp = Compartment()
    build_cache["compartment"] = [comp]

    if return_type in return_types[1:]:
        ncomps = idxs["branch_index"].value_counts().iloc[0]
        branch = Branch([comp for _ in range(ncomps)])
        build_cache["branch"] = [branch]

    if return_type in return_types[2:]:
        for cell_id, cell_groups in idxs.groupby("cell_index"):
            num_branches = cell_groups["branch_index"].nunique()
            default_parents = np.arange(num_branches) - 1  # ignores morphology
            parents = (
                default_parents if parent_branches is None else parent_branches[cell_id]
            )
            cell = Cell([branch] * num_branches, parents)
            build_cache["cell"].append(cell)

    if return_type in return_types[3:]:
        build_cache["network"] = [Network(build_cache["cell"])]

    module = build_cache[return_type][0]
    build_cache.clear()
    return module


def from_graph(
    graph: nx.DiGraph,
    ncomp: int = 4,
    max_branch_len: float = 2000.0,
    min_radius: Optional[float] = None,
    assign_groups: bool = True,
    ignore_swc_trace_errors: bool = True,
) -> Union[Network, Cell, Branch, Compartment]:
    """Build a module from a networkx graph.

    All information stored in the nodes of the graph is imported into the Module.nodes
    attribute. Edges of type "inter_branch" connect two different branches, while
    edges of type "intra_branch" connect compartments within the same branch. All other
    edges are considered synapse edges. These are added to the Module.branch_edges and
    Module.edges attributes, respectively. Additionally, the graph can contain
    global attributes, which are added as attrs, i.e. to the module instance and
    optionally can store recordings, externals, and trainables. See `to_graph` for
    how they are formatted.

    All modules that are exported to a graph from jaxley using the `to_graph` method
    can be imported back using this method. If the graph was not exported with jaxley,
    the only requirement to construct a module from a graph is that the graph contains
    edges, i.e.

    ```python
    edges = [(0,1), (1,2), (1,3)]
    graph = nx.DiGraph(edges)
    module = from_graph(graph)
    ```

    In this case, all requrired parameters are set to their default values
    and the branch structure is computed naively. However, note that in this case
    the graph has to be connected and cannot contain loops. Currently networks can
    only be imported if they have been previously exported with jaxley!

    Returns nx.Network, nx.Cell, nx.Branch, or nx.Compartment depending on
    num_cell_idxs > 1, num_branch_idxs > 1 etc.

    Possible attributes that can be read off of the graph include:
    - graph
        - ncomp: int
        - total_nbranches: int
        - cumsum_nbranches: np.ndarray
        - channels
        - synapses
        - xyzr: list[np.ndarray]
        - recordings: list[str]
        - externals: list[float]
        - trainable: dict[str, float]
    - nodes:
        - id: int (used to define groups, according to NEURON's SWC convention)
        - groups: list[str] (can also be defined directly)
        - cell_index: int
        - branch_index: int
        - comp_index: int
        - radius: float
        - length: float
        - x: float
        - y: float
        - z: float
    - edges:
        - type: str ("synapse" or "inter_branch" / "intra_branch" or None)
        - parameters: list[dict] (stores synapse parameters)

    Args:
        graph: A networkx graph representing a module.
        ncomp: The default number of segments per compartment.
            Will only be selected if the graph has not been compartmentalized yet.
        max_branch_len: Maximal length of one branch. If a branch exceeds this length,
            it is split into equal parts such that each subbranch is below
            `max_branch_len`. Will only be used if no branch structure is has been
            assigned yet.
        min_radius: If the radius of a reconstruction is below this value it is clipped.
        assign_groups: Wether to assign groups to nodes based on the the id or groups
            attribute.
        ignore_swc_trace_errors: Whether to ignore discontinuities in the swc tracing
            order. If False, this will result in split branches at these points.

    Returns:
        A module instance that is populated with the node and egde attributes of
        the nx.DiGraph."""

    ########################################
    ### Make the graph jaxley compatible ###
    ########################################

    if "type" not in graph.graph:
        try:
            graph = make_jaxley_compatible(
                graph,
                ncomp=ncomp,
                max_branch_len=max_branch_len,
                min_radius=min_radius,
                ignore_swc_trace_errors=ignore_swc_trace_errors,
            )
        except:
            raise Exception("Graph appears to be incompatible with jaxley.")

    #################################
    ### Import graph as jx.Module ###
    #################################

    # nodes and edges
    node_df = pd.DataFrame(
        [d for i, d in graph.nodes(data=True)], index=graph.nodes
    ).sort_index()
    edge_type = nx.get_edge_attributes(graph, "type")
    synapse_edges = pd.DataFrame(
        [
            {
                "pre_global_comp_index": i,
                "post_global_comp_index": j,
                **graph.edges[i, j],
            }
            for (i, j), t in edge_type.items()
            if t == "synapse"
        ]
    )

    # branches
    branch_of_node = lambda i: graph.nodes[i]["branch_index"]
    branch_edges = pd.DataFrame(
        [
            (branch_of_node(i), branch_of_node(j))
            for (i, j), t in edge_type.items()
            if t == "inter_branch"
        ],
        columns=["parent_branch_index", "child_branch_index"],
    )

    # drop special attrs from nodes and ignore error if col does not exist
    # x,y,z can be re-computed from xyzr if needed
    optional_attrs = ["recordings", "externals", "trainable", "x", "y", "z"]
    node_df = node_df.drop(columns=optional_attrs, errors="ignore")

    # synapses
    synapse_edges = synapse_edges.drop(["l", "type"], axis=1, errors="ignore")
    synapse_edges = synapse_edges.rename({"syn_type": "type"}, axis=1)
    synapse_edges.rename({"edge_index": "global_edge_index"}, axis=1, inplace=True)

    # build module
    acc_parents = []
    parent_branch_inds = branch_edges.set_index("child_branch_index").sort_index()[
        "parent_branch_index"
    ]
    for branch_inds in node_df.groupby("cell_index")["branch_index"].unique():
        root_branch_idx = branch_inds[0]
        parents = parent_branch_inds.loc[branch_inds[1:]] - root_branch_idx
        acc_parents.append([-1] + parents.tolist())

    # TODO: support inhom ncomps
    module = build_module_scaffold(node_df, graph.graph["type"], acc_parents)

    # set global attributes of module
    for k, v in graph.graph.items():
        if k not in ["type"]:
            setattr(module, k, v)

    if assign_groups and "groups" in node_df.columns:
        groups = node_df.pop("groups").explode()
        groups = (
            pd.DataFrame(groups)
            .groupby("groups")
            .apply(lambda x: x.index.values, include_groups=False)
            .to_dict()
        )
        module.groups = groups

    node_df.columns = [
        "global_" + col if "local" not in col and "index" in col else col
        for col in node_df.columns
    ]
    module.nodes[node_df.columns] = (
        node_df  # set column-wise. preserves cols not in df.
    )

    module.edges = synapse_edges if not synapse_edges.empty else module.edges

    # add all the extra attrs
    module.membrane_current_names = [c.current_name for c in module.channels]
    module.synapse_names = [s._name for s in module.synapses]

    return module


def to_graph(
    module: "jx.Module", synapses: bool = False, channels: bool = False
) -> nx.DiGraph:
    """Export the module as a networkx graph.

    Constructs a nx.DiGraph from the module. Each compartment in the module
    is represented by a node in the graph. The edges between the nodes represent
    the connections between the compartments. These edges can either be connections
    between compartments within the same branch, between different branches or
    even between different cells. In the latter case the synapse parameters
    are stored as edge attributes. Only allows one synapse per edge however!
    Additionally, global attributes of the module, for example `ncomp`, are stored as
    graph attributes.

    Exported graphs can be imported again to `jaxley` using the `from_graph` method.

    Args:
        module: A jaxley module or view instance.
        synapses: Whether to export synapses to the graph.
        channels: Whether to export ion channels to the graph.

    Returns:
        A networkx graph of the module.
    """
    module_graph = nx.DiGraph()
    module.compute_compartment_centers()  # make xyz coords attr of nodes

    # add global attrs
    module_graph.graph["type"] = module.__class__.__name__.lower()
    for attr in [
        "ncomp",
        "xyzr",
        "externals",
        "recordings",
        "trainable_params",
        "indices_set_by_trainables",
    ]:
        module_graph.graph[attr] = getattr(module, attr)

    # add nodes
    nodes = module.nodes.copy()
    nodes = nodes.drop([col for col in nodes.columns if "local" in col], axis=1)
    nodes.columns = [col.replace("global_", "") for col in nodes.columns]

    if channels:
        module_graph.graph["channels"] = module.channels
        module_graph.graph["membrane_current_names"] = [
            c.current_name for c in module.channels
        ]
    else:
        for c in module.channels:
            nodes = nodes.drop(c.name, axis=1)
            nodes = nodes.drop(list(c.channel_params), axis=1)
            nodes = nodes.drop(list(c.channel_states), axis=1)

    for col in nodes.columns:  # col wise adding preserves dtypes
        module_graph.add_nodes_from(nodes[[col]].to_dict(orient="index").items())

    nx.set_node_attributes(module_graph, [], "groups")
    if len(module.groups) > 0:
        groups_dict = {
            index: {
                "groups": [
                    key for key, value in module.groups.items() if index in value
                ]
            }
            for index in range(max(map(max, module.groups.values())) + 1)
        }
        module_graph.add_nodes_from(groups_dict.items())

    inter_branch_edges = module.branch_edges.copy()
    intra_branch_edges = []
    for i, branch_data in nodes.groupby("branch_index"):
        inds = branch_data.index.values
        intra_branch_edges += _branch_n2e(inds)

        parents = module.branch_edges["parent_branch_index"]
        children = module.branch_edges["child_branch_index"]
        inter_branch_edges.loc[parents == i, "parent_branch_index"] = inds[-1]
        inter_branch_edges.loc[children == i, "child_branch_index"] = inds[0]

    inter_branch_edges = inter_branch_edges.to_numpy()
    module_graph.add_edges_from(inter_branch_edges, type="inter_branch")
    module_graph.add_edges_from(intra_branch_edges, type="intra_branch")

    if synapses:
        syn_edges = module.edges.copy()
        multiple_syn_per_edge = syn_edges[
            ["pre_global_comp_index", "post_global_comp_index"]
        ].duplicated(keep=False)
        dupl_inds = multiple_syn_per_edge.index[multiple_syn_per_edge].values
        if multiple_syn_per_edge.any():
            warn(
                f"CAUTION: Synapses {dupl_inds} are connecting the same compartments. Exporting synapses to the graph only works if the same two compartments are connected by at most one synapse."
            )
        module_graph.graph["synapses"] = module.synapses
        module_graph.graph["synapse_param_names"] = module.synapse_param_names
        module_graph.graph["synapse_state_names"] = module.synapse_state_names
        module_graph.graph["synapse_names"] = module.synapse_names
        module_graph.graph["synapse_current_names"] = module.synapse_current_names

        syn_edges.columns = [col.replace("global_", "") for col in syn_edges.columns]
        syn_edges["syn_type"] = syn_edges["type"]
        syn_edges["type"] = "synapse"
        syn_edges = syn_edges.set_index(["pre_comp_index", "post_comp_index"])
        if not syn_edges.empty:
            for (i, j), edge_data in syn_edges.iterrows():
                module_graph.add_edge(i, j, **edge_data.to_dict())

    module_graph.graph["type"] = module.__class__.__name__.lower()

    return module_graph



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from math import pi
from typing import Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import vmap

from jaxley.utils.misc_utils import cumsum_leading_zero


def build_radiuses_from_xyzr(
    radius_fns: List[Callable],
    branch_indices: List[int],
    min_radius: Optional[float],
    ncomp: int,
) -> jnp.ndarray:
    """Return the radiuses of branches given SWC file xyzr.

    Returns an array of shape `(num_branches, ncomp)`.

    Args:
        radius_fns: Functions which, given compartment locations return the radius.
        branch_indices: The indices of the branches for which to return the radiuses.
        min_radius: If passed, the radiuses are clipped to be at least as large.
        ncomp: The number of compartments that every branch is discretized into.
    """
    # Compartment locations are at the center of the internal nodes.
    non_split = 1 / ncomp
    range_ = np.linspace(non_split / 2, 1 - non_split / 2, ncomp)

    # Build radiuses.
    radiuses = np.asarray([radius_fns[b](range_) for b in branch_indices])
    radiuses_each = radiuses.ravel(order="C")
    if min_radius is None:
        assert np.all(
            radiuses_each > 0.0
        ), "Radius 0.0 in SWC file. Set `read_swc(..., min_radius=...)`."
    else:
        radiuses_each[radiuses_each < min_radius] = min_radius

    return radiuses_each


def equal_segments(branch_property: list, ncomp_per_branch: int):
    """Generates segments where some property is the same in each segment.

    Args:
        branch_property: List of values of the property in each branch. Should have
            `len(branch_property) == num_branches`.
    """
    assert isinstance(branch_property, list), "branch_property must be a list."
    return jnp.asarray([branch_property] * ncomp_per_branch).T


def linear_segments(
    initial_val: float, endpoint_vals: list, parents: jnp.ndarray, ncomp_per_branch: int
):
    """Generates segments where some property is linearly interpolated.

    Args:
        initial_val: The value at the tip of the soma.
        endpoint_vals: The value at the endpoints of each branch.
    """
    branch_property = endpoint_vals + [initial_val]
    num_branches = len(parents)
    # Compute radiuses by linear interpolation.
    endpoint_radiuses = jnp.asarray(branch_property)

    def compute_rad(branch_ind, loc):
        start = endpoint_radiuses[parents[branch_ind]]
        end = endpoint_radiuses[branch_ind]
        return (end - start) * loc + start

    branch_inds_of_each_comp = jnp.tile(jnp.arange(num_branches), ncomp_per_branch)
    locs_of_each_comp = jnp.linspace(1, 0, ncomp_per_branch).repeat(num_branches)
    rad_of_each_comp = compute_rad(branch_inds_of_each_comp, locs_of_each_comp)

    return jnp.reshape(rad_of_each_comp, (ncomp_per_branch, num_branches)).T


def merge_cells(
    cumsum_num_branches: List[int],
    cumsum_num_branchpoints: List[int],
    arrs: List[List[np.ndarray]],
    exclude_first: bool = True,
) -> np.ndarray:
    """
    Build full list of which branches are solved in which iteration.

    From the branching pattern of single cells, this "merges" them into a single
    ordering of branches.

    Args:
        cumsum_num_branches: cumulative number of branches. E.g., for three cells with
            10, 15, and 5 branches respectively, this will should be a list containing
            `[0, 10, 25, 30]`.
        arrs: A list of a list of arrays that should be merged.
        exclude_first: If `True`, the first element of each list in `arrs` will remain
            unchanged. Useful if a `-1` (which indicates "no parent") entry should not
            be changed.

    Returns:
        A list of arrays which contain the branch indices that are computed at each
        level (i.e., iteration).
    """
    ps = []
    for i, att in enumerate(arrs):
        p = att
        if exclude_first:
            raise NotImplementedError
            p = [p[0]] + [p_in_level + cumsum_num_branches[i] for p_in_level in p[1:]]
        else:
            p = [
                p_in_level
                + np.asarray([cumsum_num_branches[i], cumsum_num_branchpoints[i]])
                for p_in_level in p
            ]
        ps.append(p)

    max_len = max([len(att) for att in arrs])
    combined_parents_in_level = []
    for i in range(max_len):
        current_ps = []
        for p in ps:
            if len(p) > i:
                current_ps.append(p[i])
        combined_parents_in_level.append(np.concatenate(current_ps))

    return combined_parents_in_level


def compute_levels(parents):
    levels = np.zeros_like(parents)

    for i, p in enumerate(parents):
        if p == -1:
            levels[i] = 0
        else:
            levels[i] = levels[p] + 1
    return levels


def compute_children_in_level(
    levels: np.ndarray, children_row_and_col: np.ndarray
) -> List[np.ndarray]:
    num_branches = len(levels)
    children_in_each_level = []
    for l in range(1, np.max(levels) + 1):
        children_in_current_level = []
        for b in range(num_branches):
            if levels[b] == l:
                children_in_current_level.append(children_row_and_col[b - 1])
        children_in_current_level = np.asarray(children_in_current_level)
        children_in_each_level.append(children_in_current_level)
    return children_in_each_level


def compute_parents_in_level(levels, par_inds, parents_row_and_col):
    level_of_parent = levels[par_inds]
    parents_in_each_level = []
    for l in range(np.max(levels)):
        parents_inds_in_current_level = np.where(level_of_parent == l)[0]
        parents_in_current_level = parents_row_and_col[parents_inds_in_current_level]
        parents_in_current_level = np.asarray(parents_in_current_level)
        parents_in_each_level.append(parents_in_current_level)
    return parents_in_each_level


def _compute_num_children(parents):
    num_branches = len(parents)
    num_children = []
    for b in range(num_branches):
        n = np.sum(np.asarray(parents) == b)
        num_children.append(n)
    return num_children


def _compute_index_of_child(parents):
    """For every branch, it returns the how many-eth child of its parent it is.

    Example:
    ```
    parents = [-1, 0, 0, 1, 1, 1]
    _compute_index_of_child(parents) -> [-1, 0, 1, 0, 1, 2]
    ```
    """
    num_branches = len(parents)
    current_num_children_for_each_branch = np.zeros((num_branches,), np.dtype("int"))
    index_of_child = [-1]
    for b in range(1, num_branches):
        index_of_child.append(current_num_children_for_each_branch[parents[b]])
        current_num_children_for_each_branch[parents[b]] += 1
    return index_of_child


def compute_children_indices(parents) -> List[jnp.ndarray]:
    """Return all children indices of every branch.

    Example:
    ```
    parents = [-1, 0, 0]
    compute_children_indices(parents) -> [[1, 2], [], []]
    ```
    """
    num_branches = len(parents)
    child_indices = []
    for b in range(num_branches):
        child_indices.append(np.where(parents == b)[0])
    return child_indices


def get_num_neighbours(
    num_children: jnp.ndarray,
    ncomp_per_branch: int,
    num_branches: int,
):
    """
    Number of neighbours of each compartment.
    """
    num_neighbours = 2 * jnp.ones((num_branches * ncomp_per_branch))
    num_neighbours = num_neighbours.at[ncomp_per_branch - 1].set(1.0)
    num_neighbours = num_neighbours.at[jnp.arange(num_branches) * ncomp_per_branch].set(
        num_children + 1.0
    )
    return num_neighbours


def local_index_of_loc(
    loc: float, global_branch_ind: int, ncomp_per_branch: int
) -> int:
    """Returns the local index of a comp given a loc [0, 1] and the index of a branch.

    This is used because we specify locations such as synapses as a value between 0 and
    1. We have to convert this onto a discrete segment here.

    Args:
        branch_ind: Index of the branch.
        loc: Location (in [0, 1]) along that branch.
        ncomp_per_branch: Number of segments of each branch.

    Returns:
        The local index of the compartment.
    """
    ncomp = ncomp_per_branch[global_branch_ind]  # only for convenience.
    possible_locs = np.linspace(0.5 / ncomp, 1 - 0.5 / ncomp, ncomp)
    ind_along_branch = np.argmin(np.abs(possible_locs - loc))
    return ind_along_branch


def loc_of_index(global_comp_index, global_branch_index, ncomp_per_branch):
    """Return location corresponding to global compartment index."""
    cumsum_ncomp = cumsum_leading_zero(ncomp_per_branch)
    index = global_comp_index - cumsum_ncomp[global_branch_index]
    ncomp = ncomp_per_branch[global_branch_index]
    return (0.5 + index) / ncomp


def compute_g_long(rad1, rad2, g_a1, g_a2, l1, l2):
    """Return the axial conductance between two compartments.

    Equations taken from `https://en.wikipedia.org/wiki/Compartmental_neuron_models`.

    The axial conductance is:
    g_long = 2 * pi * rad1^2 * rad2^2 / (l1 * r_a1 * rad2^2 + l2 * r_a2 * rad1^2)

    Here, we define `g_a = 1/r_a`, because g_a can be zero (but not infinity as this
    would be inherently unstable).
    """
    return (
        2
        * pi
        * rad1**2
        * rad2**2
        * g_a1
        * g_a2
        / (l1 * g_a2 * rad2**2 + l2 * g_a1 * rad1**2)
    )


def g_long_by_surface_area(rad1, rad2, g_a1, g_a2, l1, l2):
    """Return the voltage coupling conductance between two compartments.

    Equations taken from `https://en.wikipedia.org/wiki/Compartmental_neuron_models`.

    The axial resistivity is:
    g_long = 2 * pi * rad1^2 * rad2^2 / (l1 * r_a1 * rad2^2 + l2 * r_a2 * rad1^2)

    For voltage, we have to divide the axial conductance by the surface are of
    the sink, i.e. by A = 2 * pi * rad1 * l1

    By that, we get:
    g_axial = rad1 * rad2^2 / (l1 * r_a1 * rad2^2 + l2 * r_a2 * rad1^2) / l1

    Here, we define `g_a = 1/r_a`, because g_a can be zero (but not infinity as this
    would be inherently unstable).

    `radius`: um
    `g_a`: Siemens / cm
    `r_a`: ohm cm (unused, just for reference)
    `length_single_compartment`: um
    """
    g_long = compute_g_long(rad1, rad2, g_a1, g_a2, l1, l2)
    surface_area = 2 * pi * rad1 * l1
    return g_long / surface_area


def g_long_by_volume(rad1, rad2, g_a1, g_a2, l1, l2):
    """Return the ion diffusive constant between two compartments.

    The axial resistivity is:
    g_long = 2 * pi * rad1^2 rad2^2 / (l1 * r_a1 * rad2^2 + l2 * r_a2 * rad1^2)

    For ions, we have to divide the axial conductance by the volume of the sink,
    i.e. by V = pi * rad1^2 * l1

    This gives:
    g_axial = 2 * rad2^2 / (l1 * r_a1 * rad2^2 + l2 * r_a2 * rad1^2) / l1

    Expressed in conductances g_a (not r_a), this gives:
    g_axial = 2 * rad2^2 * g_a1 * g_a2 / (l1 * g_a2 * rad2^2 + l2 * g_a1 * rad1^2) / l1

    But here, we define `g = 1/r_a`, because g can be zero (but not infinity as this
    would be inherently unstable). In particular, one might want g=0 for ion diffusion.

    `radius`: um
    `g_a`: mM / liter / cm
    `l`: um
    """
    g_long = compute_g_long(rad1, rad2, g_a1, g_a2, l1, l2)
    volume = pi * rad1**2 * l1
    return g_long / volume


def compute_impact_on_node(rad, g_a, l):
    r"""Compute the weight with which a compartment influences its node.

    In order to satisfy Kirchhoffs current law, the current at a branch point must be
    proportional to the crosssection of the compartment. We only require proportionality
    here because the branch point equation reads:
    `g_1 * (V_1 - V_b) + g_2 * (V_2 - V_b) = 0.0`

    Because R_long = r_a * L/2 / crosssection, we get
    g_long = crosssection * 2 / L / r_a \propto rad**2 / L / r_a

    Finally, we define `g_a = 1 / r_a` (in order to allow `r_a=inf`, or `g_a=0`).

    This equation can be multiplied by any constant."""
    return rad**2 * g_a / l


def remap_to_consecutive(arr):
    """Maps an array of integers to an array of consecutive integers.

    E.g. `[0, 0, 1, 4, 4, 6, 6] -> [0, 0, 1, 2, 2, 3, 3]`
    """
    _, inverse_indices = jnp.unique(arr, return_inverse=True)
    return inverse_indices


v_interp = vmap(jnp.interp, in_axes=(None, None, 1))


def interpolate_xyzr(loc: float, coords: np.ndarray):
    """Perform a linear interpolation between xyz-coordinates.

    Args:
        loc: The location in [0,1] along the branch.
        coords: Array containing the reconstructed xyzr points of the branch.

    Return:
        Interpolated xyz coordinate at `loc`, shape `(3,).
    """
    dl = np.sqrt(np.sum(np.diff(coords[:, :3], axis=0) ** 2, axis=1))
    pathlens = np.insert(np.cumsum(dl), 0, 0)  # cummulative length of sections
    norm_pathlens = pathlens / np.maximum(1e-8, pathlens[-1])  # norm lengths to [0,1].

    return v_interp(loc, norm_pathlens, coords)


def params_to_pstate(
    params: List[Dict[str, jnp.ndarray]],
    indices_set_by_trainables: List[jnp.ndarray],
):
    """Make outputs `get_parameters()` conform with outputs of `.data_set()`.

    `make_trainable()` followed by `params=get_parameters()` does not return indices
    because these indices would also be differentiated by `jax.grad` (as soon as
    the `params` are passed to `def simulate(params)`. Therefore, in `jx.integrate`,
    we run the function to add indices to the dict. The outputs of `params_to_pstate`
    are of the same shape as the outputs of `.data_set()`."""
    return [
        {"key": list(p.keys())[0], "val": list(p.values())[0], "indices": i}
        for p, i in zip(params, indices_set_by_trainables)
    ]


def convert_point_process_to_distributed(
    current: jnp.ndarray, radius: jnp.ndarray, length: jnp.ndarray
) -> jnp.ndarray:
    """Convert current point process (nA) to distributed current (uA/cm2).

    This function gets called for synapses and for external stimuli.

    Args:
        current: Current in `nA`.
        radius: Compartment radius in `um`.
        length: Compartment length in `um`.

    Return:
        Current in `uA/cm2`.
    """
    area = 2 * pi * radius * length
    current /= area  # nA / um^2
    return current * 100_000  # Convert (nA / um^2) to (uA / cm^2)


def build_branchpoint_group_inds(
    num_branchpoints, child_belongs_to_branchpoint, start_ind_for_branchpoints
):
    branchpoint_inds_parents = start_ind_for_branchpoints + jnp.arange(num_branchpoints)
    branchpoint_inds_children = (
        start_ind_for_branchpoints + child_belongs_to_branchpoint
    )

    all_branchpoint_inds = jnp.concatenate(
        [branchpoint_inds_parents, branchpoint_inds_children]
    )
    branchpoint_group_inds = remap_to_consecutive(all_branchpoint_inds)
    return branchpoint_group_inds


def compute_morphology_indices_in_levels(
    num_branchpoints,
    child_belongs_to_branchpoint,
    par_inds,
    child_inds,
):
    """Return (row, col) to build the sparse matrix defining the voltage eqs.

    This is run at `init`, not during runtime.
    """
    branchpoint_inds_parents = jnp.arange(num_branchpoints)
    branchpoint_inds_children = child_belongs_to_branchpoint
    branch_inds_parents = par_inds
    branch_inds_children = child_inds

    children = jnp.stack([branch_inds_children, branchpoint_inds_children])
    parents = jnp.stack([branch_inds_parents, branchpoint_inds_parents])

    return {"children": children.T, "parents": parents.T}


def group_and_sum(
    values_to_sum: jnp.ndarray, inds_to_group_by: jnp.ndarray, num_branchpoints: int
) -> jnp.ndarray:
    """Group values by whether they have the same integer and sum values within group.

    This is used to construct the last diagonals at the branch points.

    Written by ChatGPT.
    """
    # Initialize an array to hold the sum of each group
    group_sums = jnp.zeros(num_branchpoints)

    # `.at[inds]` requires that `inds` is not empty, so we need an if-case here.
    # `len(inds) == 0` is the case for branches and compartments.
    if num_branchpoints > 0:
        group_sums = group_sums.at[inds_to_group_by].add(values_to_sum)

    return group_sums


def query_channel_states_and_params(d, keys, idcs):
    """Get dict with subset of keys and values from d.

    This is used to restrict a dict where every item contains __all__ states to only
    the ones that are relevant for the channel. E.g.

    ```states = {'eCa': Array([ 0.,  0., nan]}```

    will be
    ```states = {'eCa': Array([ 0.,  0.]}```

    Only loops over necessary keys, as opposed to looping over `d.items()`."""
    return dict(zip(keys, (v[idcs] for v in map(d.get, keys))))


def compute_axial_conductances(
    comp_edges: pd.DataFrame,
    params: Dict[str, jnp.ndarray],
    diffusion_states: List[str],
) -> Dict[str, jnp.ndarray]:
    """Given `comp_edges`, radius, length, r_a, cm, compute the axial conductances.

    Note that the resulting axial conductances will already by divided by the
    capacitance `cm`.
    """
    # `Compartment-to-compartment` (c2c) axial coupling conductances.
    condition = comp_edges["type"].to_numpy() == 0
    source_comp_inds = np.asarray(comp_edges[condition]["source"].to_list())
    sink_comp_inds = np.asarray(comp_edges[condition]["sink"].to_list())

    axial_conductances = jnp.stack(
        [1 / params["axial_resistivity"]]
        + [params[f"axial_diffusion_{d}"] for d in diffusion_states]
    )

    if len(sink_comp_inds) > 0:
        # For voltages, divide by the surface area.
        conds_c2c = vmap(
            vmap(g_long_by_surface_area, in_axes=(0, 0, 0, 0, 0, 0)),
            in_axes=(None, None, 0, 0, None, None),
        )(
            params["radius"][sink_comp_inds],
            params["radius"][source_comp_inds],
            axial_conductances[:1, sink_comp_inds],
            axial_conductances[:1, source_comp_inds],
            params["length"][sink_comp_inds],
            params["length"][source_comp_inds],
        )
        # .at[0] because we only divide the axial voltage conductances by the
        # capacitance, _not_ the axial conductances of the diffusing ions.
        conds_c2c = conds_c2c.at[0].divide(params["capacitance"][sink_comp_inds])
        # Multiply by 10**7 to convert (S / cm / um) -> (mS / cm^2).
        conds_c2c = conds_c2c.at[0].multiply(10**7)

        # For ion diffusion, we have to divide by the volume, not the surface area.
        conds_diffusion = vmap(
            vmap(g_long_by_volume, in_axes=(0, 0, 0, 0, 0, 0)),
            in_axes=(None, None, 0, 0, None, None),
        )(
            params["radius"][sink_comp_inds],
            params["radius"][source_comp_inds],
            axial_conductances[1:, sink_comp_inds],
            axial_conductances[1:, source_comp_inds],
            params["length"][sink_comp_inds],
            params["length"][source_comp_inds],
        )
        conds_c2c = jnp.concatenate([conds_c2c, conds_diffusion])
    else:
        conds_c2c = jnp.asarray([[]] * (len(diffusion_states) + 1))

    # `branchpoint-to-compartment` (bp2c) axial coupling conductances.
    condition = comp_edges["type"].isin([1, 2])
    sink_comp_inds = np.asarray(comp_edges[condition]["sink"].to_list())

    if len(sink_comp_inds) > 0:
        # For voltages, divide by the surface area.
        conds_bp2c = vmap(
            vmap(g_long_by_surface_area, in_axes=(0, 0, 0, 0, 0, 0)),
            in_axes=(None, None, 0, 0, None, None),
        )(
            params["radius"][sink_comp_inds],
            params["radius"][sink_comp_inds],
            axial_conductances[:1, sink_comp_inds],
            axial_conductances[:1, sink_comp_inds],
            params["length"][sink_comp_inds],
            jnp.zeros_like(params["length"][sink_comp_inds]),  # l=0 for branchpoint.
        )
        # .at[0] because we only divide the axial voltage conductances by the
        # capacitance, _not_ the axial conductances of the diffusing ions.
        conds_bp2c = conds_bp2c.at[0].divide(params["capacitance"][sink_comp_inds])
        # Multiply by 10**7 to convert (S / cm / um) -> (mS / cm^2).
        conds_bp2c = conds_bp2c.at[0].multiply(10**7)

        # For ion diffusion, we have to divide by the volume, not the surface area.
        conds_bp2c_diffusion = vmap(
            vmap(g_long_by_volume, in_axes=(0, 0, 0, 0, 0, 0)),
            in_axes=(None, None, 0, 0, None, None),
        )(
            params["radius"][sink_comp_inds],
            params["radius"][sink_comp_inds],
            axial_conductances[1:, sink_comp_inds],
            axial_conductances[1:, sink_comp_inds],
            params["length"][sink_comp_inds],
            jnp.zeros_like(params["length"][sink_comp_inds]),  # l=0 for branchpoint.
        )
        conds_bp2c = jnp.concatenate([conds_bp2c, conds_bp2c_diffusion])
    else:
        conds_bp2c = jnp.asarray([[]] * (len(diffusion_states) + 1))

    # `compartment-to-branchpoint` (c2bp) axial coupling conductances.
    condition = comp_edges["type"].isin([3, 4])
    source_comp_inds = np.asarray(comp_edges[condition]["source"].to_list())

    if len(source_comp_inds) > 0:
        conds_c2bp = vmap(
            vmap(compute_impact_on_node, in_axes=(0, 0, 0)), in_axes=(None, 0, None)
        )(
            params["radius"][source_comp_inds],
            axial_conductances[:, source_comp_inds],
            params["length"][source_comp_inds],
        )
        # For numerical stability. These values are very small, but their scale
        # does not matter.
        conds_c2bp *= 1_000
    else:
        conds_c2bp = jnp.asarray([[]] * (len(diffusion_states) + 1))

    # All axial coupling conductances.
    all_coupling_conds = jnp.concatenate([conds_c2c, conds_bp2c, conds_c2bp], axis=1)

    conds_as_dict = {}
    for i, key in enumerate(["v"] + diffusion_states):
        conds_as_dict[key] = all_coupling_conds[i]
    return conds_as_dict


def compute_children_and_parents(
    branch_edges: pd.DataFrame,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, int]:
    """Build indices used during `._init_morph_custom_spsolve()."""
    par_inds = branch_edges["parent_branch_index"].to_numpy()
    child_inds = branch_edges["child_branch_index"].to_numpy()
    child_belongs_to_branchpoint = remap_to_consecutive(par_inds)
    par_inds = np.unique(par_inds)
    return par_inds, child_inds, child_belongs_to_branchpoint



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

from typing import List

import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.synapses import IonotropicSynapse, Synapse, TanhRateSynapse, TestSynapse


def test_multiparameter_setting(SimpleNet):
    """
    Test if the correct parameters are set if one type of synapses is inserted.

    Tests global index dropping: d4daaf019596589b9430219a15f1dda0b1c34d85
    """
    net = SimpleNet(2, 1, 4)

    pre = net.cell(0).branch(0).loc(0.0)
    post = net.cell(1).branch(0).loc(0.0)
    connect(pre, post, IonotropicSynapse())

    syn_view = net.IonotropicSynapse
    syn_params = ["IonotropicSynapse_gS", "IonotropicSynapse_e_syn"]

    for p in syn_params:
        syn_view.set(p, 0.32)


def _get_synapse_view(net, synapse_name, single_idx=1, double_idxs=[2, 3]):
    """Access to the synapse view"""
    if synapse_name == "IonotropicSynapse":
        full_syn_view = net.IonotropicSynapse
        single_syn_view = net.IonotropicSynapse.edge(single_idx)
        double_syn_view = net.IonotropicSynapse.edge(double_idxs)
    if synapse_name == "TanhRateSynapse":
        full_syn_view = net.TanhRateSynapse
        single_syn_view = net.TanhRateSynapse.edge(single_idx)
        double_syn_view = net.TanhRateSynapse.edge(double_idxs)
    if synapse_name == "TestSynapse":
        full_syn_view = net.TestSynapse
        single_syn_view = net.TestSynapse.edge(single_idx)
        double_syn_view = net.TestSynapse.edge(double_idxs)
    return full_syn_view, single_syn_view, double_syn_view


@pytest.mark.parametrize(
    "synapse_type", [IonotropicSynapse, TanhRateSynapse, TestSynapse]
)
def test_set_and_querying_params_one_type(synapse_type, SimpleNet):
    """Test if the correct parameters are set if one type of synapses is inserted."""
    synapse_type = synapse_type()
    net = SimpleNet(4, 1, 4)

    for pre_ind in [0, 1]:
        for post_ind in [2, 3]:
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, synapse_type)

    # Get the synapse parameters to test setting
    syn_params = list(synapse_type.synapse_params.keys())
    for p in syn_params:
        net.set(p, 0.15)
        assert np.all(net.edges[p].to_numpy() == 0.15)

    synapse_name = type(synapse_type).__name__
    full_syn_view, single_syn_view, double_syn_view = _get_synapse_view(
        net, synapse_name
    )

    # There shouldn't be too many synapse_params otherwise this will take a long time
    for p in syn_params:
        full_syn_view.set(p, 0.32)
        assert np.all(net.edges[p].to_numpy() == 0.32)

        single_syn_view.set(p, 0.18)
        assert net.edges[p].to_numpy()[1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([0, 2, 3])] == 0.32)

        double_syn_view.set(p, 0.12)
        assert net.edges[p][0] == 0.32
        assert net.edges[p][1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([2, 3])] == 0.12)


@pytest.mark.parametrize("synapse_type", [TanhRateSynapse, TestSynapse])
def test_set_and_querying_params_two_types(synapse_type, SimpleNet):
    """Test whether the correct parameters are set."""
    synapse_type = synapse_type()
    net = SimpleNet(4, 1, 4)

    for pre_ind in [0, 1]:
        for post_ind, synapse in zip([2, 3], [IonotropicSynapse(), synapse_type]):
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, synapse)

    type1_params = list(IonotropicSynapse().synapse_params.keys())
    synapse_type_params = list(synapse_type.synapse_params.keys())

    default_synapse_type = net.edges[synapse_type_params[0]].to_numpy()[[1, 3]]

    net.set(type1_params[0], 0.15)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.15)
    if synapse_type_params[0] != type1_params[0]:
        assert np.all(
            net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == default_synapse_type
        )
    else:
        default_synapse_type = 0.15

    synapse_type_name = type(synapse_type).__name__
    synapse_type_full, synapse_type_single, synapse_type_double = _get_synapse_view(
        net, synapse_type_name, double_idxs=[0, 1]
    )

    # Generalize to all parameters
    net.IonotropicSynapse.set(type1_params[0], 0.32)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.32)
    assert np.all(
        net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == default_synapse_type
    )

    synapse_type_full.set(synapse_type_params[0], 0.18)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.32)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    net.IonotropicSynapse.edge(1).set(type1_params[0], 0.24)
    assert net.edges[type1_params[0]][0] == 0.32
    assert net.edges[type1_params[0]][2] == 0.24
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    net.IonotropicSynapse.edge([0, 1]).set(type1_params[0], 0.27)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.27)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    synapse_type_double.set(synapse_type_params[0], 0.21)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.27)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.21)


@pytest.mark.parametrize("synapse_type", [TanhRateSynapse, TestSynapse])
def test_shuffling_order_of_set(synapse_type, SimpleNet):
    """Test whether the result is the same if the order of synapses is changed."""
    synapse_type = synapse_type()

    net1 = SimpleNet(4, 1, 4)
    net2 = SimpleNet(4, 1, 4)

    connect(
        net1.cell(0).branch(0).loc(1.0),
        net1.cell(1).branch(0).loc(0.1),
        IonotropicSynapse(),
    )
    connect(
        net1.cell(1).branch(0).loc(0.6),
        net1.cell(2).branch(0).loc(0.7),
        IonotropicSynapse(),
    )
    connect(
        net1.cell(2).branch(0).loc(0.4), net1.cell(3).branch(0).loc(0.3), synapse_type
    )
    connect(
        net1.cell(3).branch(0).loc(0.1), net1.cell(1).branch(0).loc(0.1), synapse_type
    )

    # Different order as for `net1`.
    connect(
        net2.cell(3).branch(0).loc(0.1), net2.cell(1).branch(0).loc(0.1), synapse_type
    )
    connect(
        net2.cell(1).branch(0).loc(0.6),
        net2.cell(2).branch(0).loc(0.7),
        IonotropicSynapse(),
    )
    connect(
        net2.cell(2).branch(0).loc(0.4), net2.cell(3).branch(0).loc(0.3), synapse_type
    )
    connect(
        net2.cell(0).branch(0).loc(1.0),
        net2.cell(1).branch(0).loc(0.1),
        IonotropicSynapse(),
    )

    net1.insert(HH())
    net2.insert(HH())

    for i in range(4):
        net1.cell(i).branch(0).loc(0.0).record()
        net2.cell(i).branch(0).loc(0.0).record()

    voltages1 = jx.integrate(net1, t_max=5.0)
    voltages2 = jx.integrate(net2, t_max=5.0)

    assert np.max(np.abs(voltages1 - voltages2)) < 1e-8



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Sequence

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from jaxley.solver_gate import save_exp


class Transform(ABC):
    def __call__(self, x: ArrayLike) -> Array:
        return self.forward(x)

    @abstractmethod
    def forward(self, x: ArrayLike) -> Array:
        pass

    @abstractmethod
    def inverse(self, x: ArrayLike) -> Array:
        pass


class SigmoidTransform(Transform):
    """Sigmoid transformation."""

    def __init__(self, lower: ArrayLike, upper: ArrayLike) -> None:
        """This transform maps any value bijectively to the interval [lower, upper].

        Args:
            lower (ArrayLike): Lower bound of the interval.
            upper (ArrayLike): Upper bound of the interval.
        """
        super().__init__()
        self.lower = lower
        self.width = upper - lower

    def forward(self, x: ArrayLike) -> Array:
        y = 1.0 / (1.0 + save_exp(-x))
        return self.lower + self.width * y

    def inverse(self, y: ArrayLike) -> Array:
        x = (y - self.lower) / self.width
        x = -jnp.log((1.0 / x) - 1.0)
        return x


class SoftplusTransform(Transform):
    """Softplus transformation."""

    def __init__(self, lower: ArrayLike) -> None:
        """This transform maps any value bijectively to the interval [lower, inf).

        Args:
            lower (ArrayLike): Lower bound of the interval.
        """
        super().__init__()
        self.lower = lower

    def forward(self, x: ArrayLike) -> Array:
        return jnp.log1p(save_exp(x)) + self.lower

    def inverse(self, y: ArrayLike) -> Array:
        return jnp.log(save_exp(y - self.lower) - 1.0)


class NegSoftplusTransform(SoftplusTransform):
    """Negative softplus transformation."""

    def __init__(self, upper: ArrayLike) -> None:
        """This transform maps any value bijectively to the interval (-inf, upper].

        Args:
            upper (ArrayLike): Upper bound of the interval.
        """
        super().__init__(upper)

    def forward(self, x: ArrayLike) -> Array:
        return -super().forward(-x)

    def inverse(self, y: ArrayLike) -> Array:
        return -super().inverse(-y)


class AffineTransform(Transform):
    def __init__(self, scale: ArrayLike, shift: ArrayLike):
        """This transform rescales and shifts the input.

        Args:
            scale (ArrayLike): Scaling factor.
            shift (ArrayLike): Additive shift.

        Raises:
            ValueError: Scale needs to be larger than 0
        """
        if jnp.allclose(scale, 0):
            raise ValueError("a cannot be zero, must be invertible")
        self.a = scale
        self.b = shift

    def forward(self, x: ArrayLike) -> Array:
        return self.a * x + self.b

    def inverse(self, x: ArrayLike) -> Array:
        return (x - self.b) / self.a


class ChainTransform(Transform):
    """Chaining together multiple transformations"""

    def __init__(self, transforms: Sequence[Transform]) -> None:
        """A chain of transformations

        Args:
            transforms (Sequence[Transform]): Transforms to apply
        """
        super().__init__()
        self.transforms = transforms

    def forward(self, x: ArrayLike) -> Array:
        for transform in self.transforms:
            x = transform(x)
        return x

    def inverse(self, y: ArrayLike) -> Array:
        for transform in reversed(self.transforms):
            y = transform.inverse(y)
        return y


class MaskedTransform(Transform):
    def __init__(self, mask: ArrayLike, transform: Transform) -> None:
        """A masked transformation

        Args:
            mask (ArrayLike): Which elements to transform
            transform (Transform): Transformation to apply
        """
        super().__init__()
        self.mask = mask
        self.transform = transform

    def forward(self, x: ArrayLike) -> Array:
        return jnp.where(self.mask, self.transform.forward(x), x)

    def inverse(self, y: ArrayLike) -> Array:
        return jnp.where(self.mask, self.transform.inverse(y), y)


class CustomTransform(Transform):
    """Custom transformation"""

    def __init__(self, forward_fn: Callable, inverse_fn: Callable) -> None:
        """A custom transformation using a user-defined froward and
        inverse function

        Args:
            forward_fn (Callable): Forward transformation
            inverse_fn (Callable): Inverse transformation
        """
        super().__init__()
        self.forward_fn = forward_fn
        self.inverse_fn = inverse_fn

    def forward(self, x: ArrayLike) -> Array:
        return self.forward_fn(x)

    def inverse(self, y: ArrayLike) -> Array:
        return self.inverse_fn(y)


class ParamTransform:
    """Parameter transformation utility.

    This class is used to transform parameters usually from an unconstrained space to a constrained space
    and back (bacause most biophysical parameter are bounded). The user can specify a PyTree of transforms
    that are applied to the parameters.

    Attributes:
        tf_dict: A PyTree of transforms for each parameter.

    """

    def __init__(self, tf_dict: List[Dict[str, Transform]] | Transform) -> None:
        """Creates a new ParamTransform object.

        Args:
            tf_dict: A PyTree of transforms for each parameter.
        """

        self.tf_dict = tf_dict

    def forward(
        self, params: List[Dict[str, ArrayLike]] | ArrayLike
    ) -> Dict[str, Array]:
        """Pushes unconstrained parameters through a tf such that they fit the interval.

        Args:
            params: A list of dictionaries (or any PyTree) with unconstrained parameters.

        Returns:
            A list of dictionaries (or any PyTree) with transformed parameters.

        """

        return jax.tree_util.tree_map(lambda x, tf: tf.forward(x), params, self.tf_dict)

    def inverse(
        self, params: List[Dict[str, ArrayLike]] | ArrayLike
    ) -> Dict[str, Array]:
        """Takes parameters from within the interval and makes them unconstrained.

        Args:
            params: A list of dictionaries (or any PyTree) with transformed parameters.

        Returns:
            A list of dictionaries (or any PyTree) with unconstrained parameters.
        """

        return jax.tree_util.tree_map(lambda x, tf: tf.inverse(x), params, self.tf_dict)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jaxley as jx


def test_direct_distance(SimpleCell):
    ncomp = 4
    length = 15.0

    cell = SimpleCell(5, ncomp)
    cell.branch("all").loc("all").set("length", length)
    cell.compute_xyz()
    dist = cell.branch(0).loc(0.0).distance(cell.branch(0).loc(1.0))
    assert dist == (ncomp - 1) * length

    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 1])
    cell.branch("all").loc("all").set("length", length)
    cell.compute_xyz()
    dist = cell.branch(0).loc(0.0).distance(cell.branch(2).loc(1.0))
    assert dist == (3 * ncomp - 1) * length

    move_x = 220.0
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 1])
    cell.branch("all").loc("all").set("length", length)
    net = jx.Network([cell for _ in range(2)])
    net.compute_xyz()
    net.cell(1).move(move_x, 0, 0)
    dist = net.cell(1).branch(0).loc(0.0).distance(net.cell(0).branch(0).loc(0.0))
    assert dist == move_x

    dist = net.cell(0).branch(0).loc(0.0).distance(net.cell(0).branch(0).loc(0.0))
    assert dist == 0.0

    dist = net.cell(1).branch(2).loc(0.3).distance(net.cell(1).branch(2).loc(0.3))
    assert dist == 0.0

    dist = net.cell(1).branch(0).loc(0.0).distance(net.cell(1).branch(2).loc(1.0))
    assert dist == (3 * ncomp - 1) * length



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import numpy as np


def is_same_network(pre: "View", post: "View") -> bool:
    """Check if views are from the same network."""
    is_in_net = "network" in pre.base.__class__.__name__.lower()
    is_in_same_net = pre.base is post.base
    return is_in_net and is_in_same_net


def sample_comp(cell_view: "View", num: int = 1, replace=True) -> "CompartmentView":
    """Sample a compartment from a cell.

    Returns View with shape (num, num_cols)."""
    return np.random.choice(cell_view._comps_in_view, num, replace=replace)


def connect(
    pre: "View",
    post: "View",
    synapse_type: "Synapse",
):
    """Connect two compartments with a chemical synapse.

    The pre- and postsynaptic compartments must be different compartments of the
    same network.

    Args:
        pre: View of the presynaptic compartment.
        post: View of the postsynaptic compartment.
        synapse_type: The synapse to append
    """
    assert is_same_network(
        pre, post
    ), "Pre and post compartments must be part of the same network."

    pre.base._append_multiple_synapses(pre.nodes, post.nodes, synapse_type)


def fully_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
):
    """Appends multiple connections which build a fully connected layer.

    Connections are from branch 0 location 0 to a randomly chosen branch and loc.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
    """
    # Get pre- and postsynaptic cell indices.
    num_pre = len(pre_cell_view._cells_in_view)
    num_post = len(post_cell_view._cells_in_view)

    # Infer indices of (random) postsynaptic compartments.
    global_post_indices = (
        post_cell_view.nodes.groupby("global_cell_index")
        .sample(num_pre, replace=True)
        .index.to_numpy()
    )
    global_post_indices = global_post_indices.reshape((-1, num_pre), order="F").ravel()
    post_rows = post_cell_view.nodes.loc[global_post_indices]

    # Pre-synapse is at the zero-eth branch and zero-eth compartment.
    pre_rows = pre_cell_view.scope("local").branch(0).comp(0).nodes.copy()
    # Repeat rows `num_post` times. See SO 50788508.
    pre_rows = pre_rows.loc[pre_rows.index.repeat(num_post)].reset_index(drop=True)

    pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)


def sparse_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
    p: float,
):
    """Appends multiple connections which build a sparse, randomly connected layer.

    Connections are from branch 0 location 0 to a randomly chosen branch and loc.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
        p: Probability of connection.
    """
    # Get pre- and postsynaptic cell indices.
    pre_cell_inds = pre_cell_view._cells_in_view
    post_cell_inds = post_cell_view._cells_in_view
    num_pre = len(pre_cell_inds)
    num_post = len(post_cell_inds)

    num_connections = np.random.binomial(num_pre * num_post, p)
    pre_syn_neurons = np.random.choice(pre_cell_inds, size=num_connections)
    post_syn_neurons = np.random.choice(post_cell_inds, size=num_connections)

    # Sort the synapses only for convenience of inspecting `.edges`.
    sorting = np.argsort(pre_syn_neurons)
    pre_syn_neurons = pre_syn_neurons[sorting]
    post_syn_neurons = post_syn_neurons[sorting]

    # Post-synapse is a randomly chosen branch and compartment.
    global_post_indices = [
        sample_comp(post_cell_view.scope("global").cell(cell_idx))
        for cell_idx in post_syn_neurons
    ]
    global_post_indices = (
        np.hstack(global_post_indices) if len(global_post_indices) > 1 else []
    )
    post_rows = post_cell_view.base.nodes.loc[global_post_indices]

    # Pre-synapse is at the zero-eth branch and zero-eth compartment.
    global_pre_indices = pre_cell_view.base._cumsum_ncomp_per_cell[pre_syn_neurons]
    pre_rows = pre_cell_view.base.nodes.loc[global_pre_indices]

    if len(pre_rows) > 0:
        pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)


def connectivity_matrix_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
    connectivity_matrix: np.ndarray[bool],
):
    """Appends multiple connections which build a custom connected network.

    Connects pre- and postsynaptic cells according to a custom connectivity matrix.
    Entries > 0 in the matrix indicate a connection between the corresponding cells.
    Connections are from branch 0 location 0 to a randomly chosen branch and loc.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
        connectivity_matrix: A boolean matrix indicating the connections between cells.
    """
    # Get pre- and postsynaptic cell indices.
    pre_cell_inds = pre_cell_view._cells_in_view
    post_cell_inds = post_cell_view._cells_in_view
    # setting scope ensure that this works indep of current scope
    pre_nodes = pre_cell_view.scope("local").branch(0).comp(0).nodes
    pre_nodes["index"] = pre_nodes.index
    pre_cell_nodes = pre_nodes.set_index("global_cell_index")

    assert connectivity_matrix.shape == (
        len(pre_cell_inds),
        len(post_cell_inds),
    ), "Connectivity matrix must have shape (num_pre, num_post)."
    assert connectivity_matrix.dtype == bool, "Connectivity matrix must be boolean."

    # get connection pairs from connectivity matrix
    from_idx, to_idx = np.where(connectivity_matrix)
    pre_cell_inds = pre_cell_inds[from_idx]
    post_cell_inds = post_cell_inds[to_idx]

    # Sample random postsynaptic compartments (global comp indices).
    global_post_indices = np.hstack(
        [
            sample_comp(post_cell_view.scope("global").cell(cell_idx))
            for cell_idx in post_cell_inds
        ]
    )
    post_rows = post_cell_view.nodes.loc[global_post_indices]

    # Pre-synapse is at the zero-eth branch and zero-eth compartment.
    global_pre_indices = pre_cell_nodes.loc[pre_cell_inds, "index"].to_numpy()
    pre_rows = pre_cell_view.select(nodes=global_pre_indices).nodes

    pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional

import jax.numpy as jnp

from jaxley.channels import Channel
from jaxley.solver_gate import save_exp, solve_gate_exponential


class HH(Channel):
    """Hodgkin-Huxley channel."""

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True

        super().__init__(name)
        prefix = self._name
        self.channel_params = {
            f"{prefix}_gNa": 0.12,
            f"{prefix}_gK": 0.036,
            f"{prefix}_gLeak": 0.0003,
            f"{prefix}_eNa": 50.0,
            f"{prefix}_eK": -77.0,
            f"{prefix}_eLeak": -54.3,
        }
        self.channel_states = {
            f"{prefix}_m": 0.2,
            f"{prefix}_h": 0.2,
            f"{prefix}_n": 0.2,
        }
        self.current_name = f"i_HH"

    def update_states(
        self,
        states: Dict[str, jnp.ndarray],
        dt,
        v,
        params: Dict[str, jnp.ndarray],
    ):
        """Return updated HH channel state."""
        prefix = self._name
        m, h, n = states[f"{prefix}_m"], states[f"{prefix}_h"], states[f"{prefix}_n"]
        new_m = solve_gate_exponential(m, dt, *self.m_gate(v))
        new_h = solve_gate_exponential(h, dt, *self.h_gate(v))
        new_n = solve_gate_exponential(n, dt, *self.n_gate(v))
        return {f"{prefix}_m": new_m, f"{prefix}_h": new_h, f"{prefix}_n": new_n}

    def compute_current(
        self, states: Dict[str, jnp.ndarray], v, params: Dict[str, jnp.ndarray]
    ):
        """Return current through HH channels."""
        prefix = self._name
        m, h, n = states[f"{prefix}_m"], states[f"{prefix}_h"], states[f"{prefix}_n"]

        gNa = params[f"{prefix}_gNa"] * (m**3) * h  # S/cm^2
        gK = params[f"{prefix}_gK"] * n**4  # S/cm^2
        gLeak = params[f"{prefix}_gLeak"]  # S/cm^2

        return (
            gNa * (v - params[f"{prefix}_eNa"])
            + gK * (v - params[f"{prefix}_eK"])
            + gLeak * (v - params[f"{prefix}_eLeak"])
        )

    def init_state(self, states, v, params, delta_t):
        """Initialize the state such at fixed point of gate dynamics."""
        prefix = self._name
        alpha_m, beta_m = self.m_gate(v)
        alpha_h, beta_h = self.h_gate(v)
        alpha_n, beta_n = self.n_gate(v)
        return {
            f"{prefix}_m": alpha_m / (alpha_m + beta_m),
            f"{prefix}_h": alpha_h / (alpha_h + beta_h),
            f"{prefix}_n": alpha_n / (alpha_n + beta_n),
        }

    @staticmethod
    def m_gate(v):
        alpha = 0.1 * _vtrap(-(v + 40), 10)
        beta = 4.0 * save_exp(-(v + 65) / 18)
        return alpha, beta

    @staticmethod
    def h_gate(v):
        alpha = 0.07 * save_exp(-(v + 65) / 20)
        beta = 1.0 / (save_exp(-(v + 35) / 10) + 1)
        return alpha, beta

    @staticmethod
    def n_gate(v):
        alpha = 0.01 * _vtrap(-(v + 55), 10)
        beta = 0.125 * save_exp(-(v + 65) / 80)
        return alpha, beta


def _vtrap(x, y):
    return x / (save_exp(x / y) - 1.0)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.pumps.pump import Pump  # isort: skip
from jaxley.pumps.ca_pump import CaPump
from jaxley.pumps.faraday_electrolysis import CaFaradayConcentrationChange
from jaxley.pumps.nernstreversal import CaNernstReversal



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Any, Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp


class TypeOptimizer:
    """`optax` wrapper which allows different argument values for different params."""

    def __init__(
        self,
        optimizer: Callable,
        optimizer_args: Dict[str, Any],
        opt_params: List[Dict[str, jnp.ndarray]],
    ):
        """Create the optimizers.

        This requires access to `opt_params` in order to know how many optimizers
        should be created. It creates `len(opt_params)` optimizers.

        Example usage:
        ```
        lrs = {"HH_gNa": 0.01, "radius": 1.0}
        optimizer = TypeOptimizer(lambda lr: optax.adam(lr), lrs, opt_params)
        opt_state = optimizer.init(opt_params)
        ```

        ```
        optimizer_args = {"HH_gNa": [0.01, 0.4], "radius": [1.0, 0.8]}
        optimizer = TypeOptimizer(
            lambda args: optax.sgd(args[0], momentum=args[1]),
            optimizer_args,
            opt_params
        )
        opt_state = optimizer.init(opt_params)
        ```

        Args:
            optimizer: A Callable that takes the learning rate and returns the
                `optax.optimizer` which should be used.
            optimizer_args: The arguments for different kinds of parameters.
                Each item of the dictionary will be passed to the `Callable` passed to
                `optimizer`.
            opt_params: The parameters to be optimized. The exact values are not used,
                only the number of elements in the list and the key of each dict.
        """
        self.base_optimizer = optimizer

        self.optimizers = []
        for params in opt_params:
            names = list(params.keys())
            assert len(names) == 1, "Multiple parameters were added at once."
            name = names[0]
            optimizer = self.base_optimizer(optimizer_args[name])
            self.optimizers.append({name: optimizer})

    def init(self, opt_params: List[Dict[str, jnp.ndarray]]) -> List:
        """Initialize the optimizers. Equivalent to `optax.optimizers.init()`."""
        opt_states = []
        for params, optimizer in zip(opt_params, self.optimizers):
            name = list(optimizer.keys())[0]
            opt_state = optimizer[name].init(params)
            opt_states.append(opt_state)
        return opt_states

    def update(self, gradient: jnp.ndarray, opt_state: List) -> Tuple[List, List]:
        """Update the optimizers. Equivalent to `optax.optimizers.update()`."""
        all_updates = []
        new_opt_states = []
        for grad, state, opt in zip(gradient, opt_state, self.optimizers):
            name = list(opt.keys())[0]
            updates, new_opt_state = opt[name].update(grad, state)
            all_updates.append(updates)
            new_opt_states.append(new_opt_state)
        return all_updates, new_opt_states
